/*******************************************************************************/
/* Robotarm model                                                              */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* if you reuse this code, please cite                                         */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef struct state {
  int mode;
  int phase;
  int reset;
  double speed;
  double pos;
} t_state;

#define POS_INIT 0.0
#define POS_MAX 3.0
#define POS_MIN -3.0
#define A 0.25 //acceleration
#define V 0.5 //max speed

#define I_UP 1
#define I_NONE 0
#define I_DOWN -1

#define M_IDLE 0
#define M_UP 1
#define M_DOWN -1

#define P_CST 0
#define P_ACC 1
#define P_DEC -1

typedef struct io {
  int joystick; 
  int reset;
} t_io;

void init(t_state *s) {
  s->mode = M_IDLE;
  s->phase = P_CST;
  s->reset = 0;
  s->pos = POS_INIT;
  s->speed = 0.0;
}

void compute(t_io* i, t_state *s) {
  double abs_speed = s->speed;
  if(s->speed<0) abs_speed = -s->speed;
  double dist_est = A*(abs_speed/A)*(abs_speed/A)*1.0; //(safely?) estimate braking distance
  if(i->reset) { s->reset = 1; }
  if(s->mode==M_IDLE) {
    if(s->reset) {
      if(s->pos>POS_INIT) { s->mode = M_DOWN; s->phase = P_ACC; }
      else if(s->pos<POS_INIT) { s->mode = M_UP; s->phase = P_ACC; }
      else s->reset = 0;
    }
    else if(i->joystick==I_UP) {
      s->mode = M_UP;
      s->phase = P_ACC;
    }
    else if(i->joystick==I_DOWN) {
      s->mode = M_DOWN;
      s->phase = P_ACC;
    }
  }
  else if(s->mode==M_UP) {
    if(s->phase==P_CST) {
      if(!s->reset && (i->joystick==I_DOWN || s->pos>=POS_MAX-dist_est) ||
        s->reset && s->pos>=POS_INIT-dist_est) { s->phase = P_DEC; }
    }
    else if(s->phase==P_ACC) {
      if(!s->reset && (i->joystick==I_DOWN || s->pos>=POS_MAX-dist_est) ||
        s->reset && s->pos>=POS_INIT-dist_est) { s->phase = P_DEC; }
      else if(s->speed<V) s->speed += A;
      else { s->phase = P_CST; }
    }
    else if(s->phase==P_DEC) {
      if(s->speed>0) s->speed -= A;
      else { 
        if(!s->reset && i->joystick==I_DOWN) { s->mode = M_DOWN; s->phase = P_ACC; }
        else { s->mode = M_IDLE; s->phase = P_CST; }
      }
    }
  }
  else if(s->mode==M_DOWN) {
    if(s->phase==P_CST) {
      if(!s->reset && (i->joystick==I_UP || s->pos<=POS_MIN+dist_est) ||
        s->reset && s->pos<=POS_INIT+dist_est) { s->phase = P_DEC; }
    }
    else if(s->phase==P_ACC) {
      if(!s->reset && (i->joystick==I_UP || s->pos<=POS_MIN+dist_est) ||
        s->reset && s->pos<=POS_INIT+dist_est) { s->phase = P_DEC; }
      else if(s->speed>-V) s->speed -= A;
      else { s->phase = P_CST; }
    }
    else if(s->phase==P_DEC) {
      if(s->speed<0) s->speed += A;
      else { 
        if(i->joystick==I_UP) { s->mode = M_UP; s->phase = P_ACC; }
        else { s->mode = M_IDLE; s->phase = P_CST; }
      }
    }
  }
  s->pos += s->speed;
}

#ifndef NO_CBMC

void input_assume(t_io* i, t_state* s) {
 __CPROVER_assume(
    -1<=i->joystick && i->joystick<=1 &&
    0<=i->reset && i->reset<=1
  );
}

extern t_io nondet_input();

int main(int argc, char** argv) 
{
  t_state s;
  init(&s);
  
  while(1)
  {
    //step
    t_io i = nondet_input();
    input_assume(&i,&s);
    t_state s_old = s;
    compute(&i,&s);

    //properties
    if(s_old.mode==M_IDLE && !s_old.reset && !i.reset && i.joystick==I_UP)
      __CPROVER_assert(s.mode==M_UP && s.phase==P_ACC,"c::prop1");

    if(s_old.mode==M_IDLE && !s_old.reset && !i.reset && i.joystick==I_DOWN)
      __CPROVER_assert(s.mode==M_DOWN && s.phase==P_ACC,"c::prop2");

    if(s_old.mode==M_IDLE && s_old.pos<POS_INIT && i.reset)
      __CPROVER_assert(s.reset && s.mode==M_UP,"c::prop3");

    if(s_old.mode==M_IDLE && s_old.pos>POS_INIT && i.reset)
      __CPROVER_assert(s.reset && s.mode==M_DOWN,"c::prop4");

    if(i.reset)
      __CPROVER_assert(s.reset,"c::prop5");

    if(s_old.phase==P_ACC && (s_old.speed==V || s_old.speed==-V) &&
		   i.joystick==I_NONE && !i.reset)
      __CPROVER_assert(s.phase==P_CST,"c::prop6");

    if(s_old.mode==M_UP && !s_old.reset && !i.reset && i.joystick==I_DOWN)
      __CPROVER_assert(s.mode==M_UP && s.phase==P_DEC,"c::prop7");

    if(s_old.mode==M_DOWN && !s_old.reset && !i.reset && i.joystick==I_UP)
      __CPROVER_assert(s.mode==M_DOWN && s.phase==P_DEC,"c::prop8");

    if(s_old.reset && (i.joystick==I_UP || i.joystick==I_DOWN))
      __CPROVER_assert(s.reset,"c::prop9");

    if(s_old.mode==M_IDLE && !s_old.reset && !i.reset && 
       (i.joystick==I_UP || i.joystick==I_DOWN))
      __CPROVER_assert(s.mode!=M_IDLE && s.phase==P_ACC,"c::prop10");

    if(s_old.mode==M_UP && i.joystick==I_UP)
      __CPROVER_assert(s.mode==M_UP,"c::prop11");

    if(s_old.mode==M_DOWN && i.joystick==I_DOWN)
      __CPROVER_assert(s.mode==M_DOWN,"c::prop12");

    if(s_old.phase==P_DEC && s_old.speed==0.0 &&
       i.joystick==I_NONE && !i.reset)
      __CPROVER_assert(s.mode==M_IDLE && s.phase==P_CST,"c::prop13");

    if(s_old.mode==M_UP && s_old.phase==P_ACC && s_old.speed==V &&
		   i.joystick==I_NONE && !i.reset)
     __CPROVER_assert(s.mode==M_UP && s.phase==P_CST && s.speed==V,"c::prop14");

    if(s_old.mode==M_DOWN && s_old.phase==P_ACC && s_old.speed==-V &&
		   i.joystick==I_NONE && !i.reset)
      __CPROVER_assert(s.mode==M_DOWN && 
		       s.phase==P_CST && s.speed==-V,"c::prop15");

    if(s_old.mode==M_IDLE && i.joystick==I_NONE && !i.reset)
      __CPROVER_assert(s.mode==M_IDLE && s.speed==0.0,"c::prop16");

    if(s_old.mode==M_IDLE && s_old.reset && s_old.pos==POS_INIT)
      __CPROVER_assert(!s.reset,"c::prop17");

    if(s_old.pos==POS_MAX)
      __CPROVER_assert(s.speed<=0,"c::prop18");

    if(s_old.pos==POS_MIN)
      __CPROVER_assert(s.speed>=0,"c::prop19");


    //final
    if(s.mode==M_IDLE && !s.reset && !i.reset && i.joystick==I_UP) 
      __CPROVER_assert(1,"c::final");
   }

  return 0;
}

#endif
