/*******************************************************************************/
/* Car window model (extended with error state)                                */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     http://www.mathworks.co.uk/products/simulink/examples.html              */
/* if you reuse this code, please cite                                         */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define FAILSAFE_TIMEOUT 5
#define END_TIMEOUT 2
#define EM_TIMEOUT 2

typedef struct state {
  int stop;
  int ready;
  int posbottom;
  int postop;
  int posnotsure;

  int move;
  int up;
  int cont;

  int emergency;
  int endstop;

  int error;

  int timer;
} t_state;

typedef struct io {
  int ready; 
  int up;
  int down; 
  int nearend; 
  int obstacle; 
  int error; 
} t_io;

void init(t_state *s) {
  s->stop = 1;
  s->ready = 0;
  s->posbottom = 0;
  s->postop = 0;
  s->posnotsure = 1;
  s->move = 0;
  s->up = 0;
  s->cont = 0;
  s->emergency = 0;
  s->endstop = 0;
  s->error = 0;
  s->timer = 0;
}

void compute(t_io* i, t_state *s) {
  if(s->error) return;
  if(s->ready && i->error) { s->error = 1; s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
  if(s->stop) {
    if(!s->ready) {
      if(i->ready) { s->ready = 1; return; }
    }
    else { //ready
      if(s->postop && !s->posbottom && !s->posnotsure) {
        if(i->down && !i->up) { s->stop = 0; s->move = 1; s->up = 0; s->posnotsure = 1; s->postop = 0; return; }
      }
      else if(!s->postop && s->posbottom && !s->posnotsure) {
        if(i->up && !i->down) { s->stop = 0; s->move = 1; s->up = 1; s->posnotsure = 1; s->posbottom = 0; return; }
      }
      else if(!s->postop && !s->posbottom && s->posnotsure) {
        if(i->up && !i->down) { s->stop = 0; s->move = 1; s->up = 1; return; }
        if(i->down && !i->up) { s->stop = 0; s->move = 1; s->up = 0; return; }
      }
    }
  }
  else if(s->move) {
    s->timer++;
    if(s->timer>=FAILSAFE_TIMEOUT+1) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
    if(i->nearend) { s->endstop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
    if(i->obstacle) { s->emergency = 1; s->move = 0; s->cont = 0; s->up = 0; s->timer = 0; return; }
    if(s->up) {
      if(!s->cont) {
        if(!i->up) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
        if(i->up) { s->cont = 1; return; }
      }
      else { //cont
        if(i->down) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
      }
    }
    else { //down
      if(!s->cont) {
        if(!i->down) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
        if(i->down) { s->cont = 1; return; }
      }
      else { //cont
        if(i->up) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
      }
    }
  }
  else if(s->emergency) {
    s->timer++;
    if(s->timer>=EM_TIMEOUT+1) { s->stop = 1; s->emergency = 0; s->timer = 0; return; }
  }
  else if(s->endstop) {
    s->timer++;
    if(s->up) {
      if(s->timer>=END_TIMEOUT+1) { s->stop = 1; s->endstop = 0; s->postop = 1; s->posnotsure = 0; s->timer = 0; return; }
    }
    else { //down
      if(s->timer>=END_TIMEOUT+1) { s->stop = 1; s->endstop = 0; s->posbottom = 1; s->posnotsure = 0; s->timer = 0; return; }
    }
  }
}

#ifndef NO_CBMC

void input_assume(t_io* i, t_state* s) {
  __CPROVER_assume((0<=i->ready) && (i->ready<=1) &&
                   (0<=i->up) && (i->up<=1) &&
                   (0<=i->down) && (i->down<=1) &&
                   (0<=i->nearend) && (i->nearend<=1) &&
                   (0<=i->obstacle) && (i->obstacle<=1) &&
                   (0<=i->error) && (i->error<=1) &&
                   (!i->up && !i->down || i->up && !i->down || !i->up && i->down)
 );
}

extern t_io nondet_input();

int main(int argc, char** argv) 
{
  t_state s;
  init(&s);
  
  while(1)
  {
    //step
    t_io i = nondet_input();
    input_assume(&i,&s);
    t_state s_old = s;
    compute(&i,&s);

    //properties
    if(!s_old.error && s_old.stop && s_old.ready && 
       (!s_old.posbottom && i.down || !s_old.postop && i.up))
      __CPROVER_assert(s.move,"c::prop1");

    if(!s_old.error && s_old.move && i.obstacle)
      __CPROVER_assert(s.emergency,"c::prop2");

    if(!s_old.error && s_old.emergency && s_old.timer>=EM_TIMEOUT)
      __CPROVER_assert(s.stop,"c::prop3");

    if(!s_old.error && s_old.endstop && s_old.timer>=END_TIMEOUT)
      __CPROVER_assert(s.stop,"c::prop4");

    if(!s_old.error && s_old.move && i.nearend)
      __CPROVER_assert(s.endstop,"c::prop5");

    if(!s_old.error && s_old.move && s_old.cont && 
       (s_old.up && i.down || !s_old.up && i.up))
      __CPROVER_assert(s.stop,"c::prop6");

    if(!s_old.error && s_old.move && s_old.timer>=FAILSAFE_TIMEOUT)
      __CPROVER_assert(s.stop,"c::prop7");

    if(!s_old.error && s_old.stop && s_old.ready && !s_old.postop && i.up)
      __CPROVER_assert(s.move,"c::prop8");

    if(!s_old.error && s_old.stop && s_old.ready && !s_old.posbottom && i.down)
      __CPROVER_assert(s.move,"c::prop9");

    if(!s_old.error && s_old.move && s_old.cont && s_old.up && i.down)
      __CPROVER_assert(s.stop,"c::prop10");

    if(!s_old.error && s_old.move && s_old.cont && !s_old.up && i.up)
      __CPROVER_assert(s.stop,"c::prop11");

    if(!s_old.error && s_old.endstop && s_old.up && s_old.timer>=END_TIMEOUT);
      __CPROVER_assert(s.postop,"c::prop12");

    if(!s_old.error && s_old.endstop && !s_old.up && s_old.timer>=END_TIMEOUT)
      __CPROVER_assert(s.posbottom,"c::prop13");

    if(!s_old.error && s_old.move && s_old.timer>=FAILSAFE_TIMEOUT)
      __CPROVER_assert(s.posnotsure,"c::prop14");

    if(!s_old.error && s_old.stop && s_old.ready && s_old.postop && i.up)
      __CPROVER_assert(s.stop,"c::prop15");

    if(!s_old.error && s_old.stop && s_old.ready && s_old.posbottom && i.down)
      __CPROVER_assert(s.stop,"c::prop16");

    if(!s_old.error && !s_old.ready && !i.ready)
      __CPROVER_assert(!s.ready,"c::prop17");

    if(!s_old.error && s_old.move && !s_old.cont && s_old.up && 
       s_old.timer<FAILSAFE_TIMEOUT && i.up && !i.obstacle && !i.nearend)
      __CPROVER_assert(s.cont,"c::prop18");

    if(!s_old.error && s_old.move && !s_old.cont && !s_old.up && 
       s_old.timer<FAILSAFE_TIMEOUT && i.down && !i.obstacle && !i.nearend)
      __CPROVER_assert(s.cont,"c::prop19");

    if(!s_old.error && s_old.move && !s_old.cont && 
       s_old.timer<FAILSAFE_TIMEOUT && 
       (!s_old.up && i.down || s_old.up && i.up) && 
       !i.obstacle && !i.nearend)
      __CPROVER_assert(s.cont,"c::prop20");

    if(s_old.ready && !s_old.stop && !s_old.error && i.error)
      __CPROVER_assert(s.error && s.stop,"c::prop21");

    if(s_old.ready && s_old.stop && !s_old.error && i.error)
      __CPROVER_assert(s.error && s.stop,"c::prop22");

    if(s_old.ready && s_old.error && (i.up || i.down))
      __CPROVER_assert(s.error && s.stop,"c::prop23");

    if(!s_old.ready && i.ready)
      __CPROVER_assert(s.ready,"c::prop24");

    if(s_old.move && i.nearend)
      __CPROVER_assert(s.endstop,"c::prop25");

    if(s_old.move && s_old.cont && (s_old.up && i.down || !s_old.up && i.up))
      __CPROVER_assert(s.stop,"c::prop26");

    if(s_old.move && s_old.timer>=FAILSAFE_TIMEOUT)
      __CPROVER_assert(s.stop,"c::prop27");

    //Final
    if(s.stop && s.ready && !s.move && 
       !s.emergency && !s.endstop || s.error) 
      __CPROVER_assert(1,"c::final");
  }

  return 0;
}

#endif
