/*******************************************************************************/
/* Car window model (extended with error state)                                */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     http://www.mathworks.co.uk/products/simulink/examples.html              */
/* if you reuse this code, please cite                                         */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define FAILSAFE_TIMEOUT 5
#define END_TIMEOUT 2
#define EM_TIMEOUT 2

typedef struct state {
  int stop;
  int ready;
  int posbottom;
  int postop;
  int posnotsure;

  int move;
  int up;
  int cont;

  int emergency;
  int endstop;

  int error;

  int timer;
} t_state;

typedef struct io {
  int ready; 
  int up;
  int down; 
  int nearend; 
  int obstacle; 
  int error; 
} t_io;

void init(t_state *s) {
  s->stop = 1;
  s->ready = 0;
  s->posbottom = 0;
  s->postop = 0;
  s->posnotsure = 1;
  s->move = 0;
  s->up = 0;
  s->cont = 0;
  s->emergency = 0;
  s->endstop = 0;
  s->error = 0;
  s->timer = 0;
}

void compute(t_io* i, t_state *s) {
  if(s->error) return;
  if(s->ready && i->error) { s->error = 1; s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
  if(s->stop) {
    if(!s->ready) {
      if(i->ready) { s->ready = 1; return; }
    }
    else { //ready
      if(s->postop && !s->posbottom && !s->posnotsure) {
        if(i->down && !i->up) { s->stop = 0; s->move = 1; s->up = 0; s->posnotsure = 1; s->postop = 0; return; }
      }
      else if(!s->postop && s->posbottom && !s->posnotsure) {
        if(i->up && !i->down) { s->stop = 0; s->move = 1; s->up = 1; s->posnotsure = 1; s->posbottom = 0; return; }
      }
      else if(!s->postop && !s->posbottom && s->posnotsure) {
        if(i->up && !i->down) { s->stop = 0; s->move = 1; s->up = 1; return; }
        if(i->down && !i->up) { s->stop = 0; s->move = 1; s->up = 0; return; }
      }
    }
  }
  else if(s->move) {
    s->timer++;
    if(s->timer>=FAILSAFE_TIMEOUT+1) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
    if(i->nearend) { s->endstop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
    if(i->obstacle) { s->emergency = 1; s->move = 0; s->cont = 0; s->up = 0; s->timer = 0; return; }
    if(s->up) {
      if(!s->cont) {
        if(!i->up) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
        if(i->up) { s->cont = 1; return; }
      }
      else { //cont
        if(i->down) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
      }
    }
    else { //down
      if(!s->cont) {
        if(!i->down) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
        if(i->down) { s->cont = 1; return; }
      }
      else { //cont
        if(i->up) { s->stop = 1; s->move = 0; s->cont = 0; s->timer = 0; return; }
      }
    }
  }
  else if(s->emergency) {
    s->timer++;
    if(s->timer>=EM_TIMEOUT+1) { s->stop = 1; s->emergency = 0; s->timer = 0; return; }
  }
  else if(s->endstop) {
    s->timer++;
    if(s->up) {
      if(s->timer>=END_TIMEOUT+1) { s->stop = 1; s->endstop = 0; s->postop = 1; s->posnotsure = 0; s->timer = 0; return; }
    }
    else { //down
      if(s->timer>=END_TIMEOUT+1) { s->stop = 1; s->endstop = 0; s->posbottom = 1; s->posnotsure = 0; s->timer = 0; return; }
    }
  }
}

#ifndef NO_CBMC

void state_assume(t_state* s) {
  __CPROVER_assume(
    0<=s->error && s->error<=1 &&
    0<=s->stop && s->stop<=1 &&
    0<=s->move && s->move<=1 &&
    0<=s->emergency && s->emergency<=1 &&
    0<=s->endstop && s->endstop<=1 &&
    0<=s->ready && s->ready<=1 &&
    0<=s->up && s->up<=1 &&
    0<=s->cont && s->cont<=1 &&
    0<=s->postop && s->postop<=1 &&
    0<=s->posbottom && s->posbottom<=1 &&
    0<=s->posnotsure && s->posnotsure<=1 &&
    (
      s->stop && !s->move && !s->emergency && !s->endstop ||
      !s->stop && s->move && !s->emergency && !s->endstop ||
      !s->stop && !s->move && s->emergency && !s->endstop ||
      !s->stop && !s->move && !s->emergency && s->endstop 
    ) 
    && (s->move || !s->cont) 
    && (s->stop || s->ready) 
   );
}

void input_assume(t_io* i) {
  __CPROVER_assume((0<=i->ready) && (i->ready<=1) &&
                   (0<=i->up) && (i->up<=1) &&
                   (0<=i->down) && (i->down<=1) &&
                   (0<=i->nearend) && (i->nearend<=1) &&
                   (0<=i->obstacle) && (i->obstacle<=1) &&
                   (0<=i->error) && (i->error<=1) &&
                   (!i->up && !i->down || i->up && !i->down || !i->up && i->down)
 );
}

void final(t_io* i, t_state* s) {
  __CPROVER_assume(s->stop && s->ready && !s->move && !s->emergency && !s->endstop || s->error);
}

void prop1(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->stop && s->ready && (!s->posbottom && i->down || !s->postop && i->up));
  //step
  compute(i,s);
  //postcondition
  assert(s->move);
}

void prop2(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && i->obstacle);
  //step
  compute(i,s);
  //postcondition
  assert(s->emergency);
}

void prop3(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->emergency && s->timer>=EM_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop4(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->endstop && s->timer>=END_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop5(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && i->nearend);
  //step
  compute(i,s);
  //postcondition
  assert(s->endstop);
}

void prop6(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && s->cont && (s->up && i->down || !s->up && i->up));
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop7(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && s->timer>=FAILSAFE_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop8(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->stop && s->ready && !s->postop && i->up);
  //step
  compute(i,s);
  //postcondition
  assert(s->move);
}

void prop9(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->stop && s->ready && !s->posbottom && i->down);
  //step
  compute(i,s);
  //postcondition
  assert(s->move);
}

void prop10(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && s->cont && s->up && i->down);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop11(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && s->cont && !s->up && i->up);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop12(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->endstop && s->up && s->timer>=END_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->postop);
}

void prop13(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->endstop && !s->up && s->timer>=END_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->posbottom);
}

void prop14(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && s->timer>=FAILSAFE_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->posnotsure);
}

void prop15(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->stop && s->ready && s->postop && i->up);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop16(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->stop && s->ready && s->posbottom && i->down);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop17(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && !s->ready && !i->ready);
  //step
  compute(i,s);
  //postcondition
  assert(!s->ready);
}

void prop18(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && !s->cont && s->up && s->timer<FAILSAFE_TIMEOUT && i->up && !i->obstacle && !i->nearend);
  //step
  compute(i,s);
  //postcondition
  assert(s->cont);
}

void prop19(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && !s->cont && !s->up && s->timer<FAILSAFE_TIMEOUT && i->down && !i->obstacle && !i->nearend);
  //step
  compute(i,s);
  //postcondition
  assert(s->cont);
}

void prop20(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->error && s->move && !s->cont && s->timer<FAILSAFE_TIMEOUT && 
		   (!s->up && i->down || s->up && i->up) && 
                   !i->obstacle && !i->nearend);
  //step
  compute(i,s);
  //postcondition
  assert(s->cont);
}

void prop21(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->ready && !s->stop && !s->error && i->error);
  //step
  compute(i,s);
  //postcondition
  assert(s->error && s->stop);
}

void prop22(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->ready && s->stop && !s->error && i->error);
  //step
  compute(i,s);
  //postcondition
  assert(s->error && s->stop);
}

void prop23(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->ready && s->error && (i->up || i->down));
  //step
  compute(i,s);
  //postcondition
  assert(s->error && s->stop);
}

void prop24(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(!s->ready && i->ready);
  //step
  compute(i,s);
  //postcondition
  assert(s->ready);
}

void prop25(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->move && i->nearend);
  //step
  compute(i,s);
  //postcondition
  assert(s->endstop);
}

void prop26(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->move && s->cont && (s->up && i->down || !s->up && i->up));
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void prop27(t_io *i, t_state* s) {
  //precondition
  input_assume(i);
  __CPROVER_assume(s->move && s->timer>=FAILSAFE_TIMEOUT);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop);
}

void main() {}

#endif
