/*******************************************************************\

Module: Main Module 

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <iostream>
#include <fstream>
#include <memory>
#include <cstdlib>

#include <util/config.h>
#include <util/expr_util.h>
#include <util/language.h>

#include <util/time_stopping.h>

#include <goto-programs/goto_convert_functions.h>
#include <goto-programs/remove_function_pointers.h>
#include <goto-programs/goto_inline.h>
#include <goto-programs/show_properties.h>
#include <goto-programs/set_properties.h>
#include <goto-programs/read_goto_binary.h>
#include <goto-programs/interpreter.h>
#include <goto-programs/string_abstraction.h>
#include <goto-programs/string_instrumentation.h>
#include <goto-programs/loop_ids.h>
#include <goto-programs/link_to_library.h>
#include <util/rename.h>

#include <pointer-analysis/goto_program_dereference.h>
#include <pointer-analysis/add_failed_symbols.h>

#include <langapi/mode.h>

#include "chaincover_parseoptions.h"
#include "cbmc_solvers.h"
#include "bmc.h"
#include "version.h"
#include "xml_interface.h"

//#include "graph.h"
//#include "reach_bmc.h"
#include "testvector.h"
#include "config_parser_xml.h"

#include "test_sat/test_sat_incremental.h"
#include "test_sat/test_sat_nonincremental.h"

#include "testchain_opt/testchain_opt.h"
#include "testchain_sat/testchain_sat.h"

//#include "mcdc.h"
#include <goto-instrument/branchtest.h>
//#include <goto-instrument/accelerate/accelerate.h>

#define RM_CMD "rm -f"

/*******************************************************************\

Function: chaincover_parseoptionst::cbmc_parseoptionst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

chaincover_parseoptionst::chaincover_parseoptionst(int argc, const char **argv):
  parseoptions_baset(CBMC_OPTIONS, argc, argv),
  xml_interfacet(cmdline),
  language_uit("ChainCover " CHAINCOVER_VERSION, cmdline)
{
}
  
/*******************************************************************\

Function: chaincover_parseoptionst::cbmc_parseoptionst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

::chaincover_parseoptionst::chaincover_parseoptionst(
  int argc,
  const char **argv,
  const std::string &extra_options):
  parseoptions_baset(CBMC_OPTIONS+extra_options, argc, argv),
  xml_interfacet(cmdline),
  language_uit("ChainCover " CHAINCOVER_VERSION, cmdline)
{
}

/*******************************************************************\

Function: chaincover_parseoptionst::eval_verbosity

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void chaincover_parseoptionst::eval_verbosity()
{
  // this is our default verbosity
  int v=messaget::M_STATISTICS;
  
  if(cmdline.isset("verbosity"))
  {
    v=unsafe_string2int(cmdline.getval("verbosity"));
    if(v<0)
      v=0;
    else if(v>10)
      v=10;
  }
  
  ui_message_handler.set_verbosity(v);
}


/*******************************************************************\

Function: chaincover_parseoptionst::get_command_line_options

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void chaincover_parseoptionst::get_command_line_options(optionst &options)
{
  if(config.set(cmdline)) {
    usage_error();
    exit(1);
  }

  if(!cmdline.isset("config") && 
     (cmdline.isset("chain") || cmdline.isset("chain-properties"))) {
    status("No configuration given: inferring configuration from source. Use --output-config to write configuration file to disk.");
  }

  if(!cmdline.isset("config") && cmdline.isset("output-config")) {
    throw std::string("Use --config to specify configuration file to write.");
  }

  if(cmdline.isset("do-not-minimize"))
    options.set_option("minimize-test-suite", false);
  else
    options.set_option("minimize-test-suite", true);

  if(cmdline.isset("program-only"))
    options.set_option("program-only", true);

  if(cmdline.isset("show-vcc"))
    options.set_option("show-vcc", true);

  if(cmdline.isset("cover-assertions"))
    options.set_option("cover-assertions", true);

  if(cmdline.isset("no-simplify"))
    options.set_option("simplify", false);
  else
    options.set_option("simplify", true);

  if(cmdline.isset("unwind-max"))
    options.set_option("unwind-max", cmdline.getval("unwind-max"));
  if(cmdline.isset("unwind"))
    options.set_option("unwind", cmdline.getval("unwind"));

  if(cmdline.isset("depth"))
    options.set_option("depth", cmdline.getval("depth"));

  if(cmdline.isset("debug-level"))
    options.set_option("debug-level", cmdline.getval("debug-level"));

  if(cmdline.isset("slice-by-trace"))
    options.set_option("slice-by-trace", cmdline.getval("slice-by-trace"));

  if(cmdline.isset("unwindset"))
    options.set_option("unwindset", cmdline.getval("unwindset"));

  if(cmdline.isset("incremental-check"))
    options.set_option("incremental-check", cmdline.getval("incremental-check"));

  // constant propagation
  if(cmdline.isset("no-propagation"))
    options.set_option("propagation", false);
  else
    options.set_option("propagation", true);

  // check array bounds
  if(cmdline.isset("bounds-check"))
    options.set_option("bounds-check", true);
  else
    options.set_option("bounds-check", false);

  // check division by zero
  if(cmdline.isset("div-by-zero-check"))
    options.set_option("div-by-zero-check", true);
  else
    options.set_option("div-by-zero-check", false);

  // check overflow/underflow
  if(cmdline.isset("signed-overflow-check"))
    options.set_option("signed-overflow-check", true);
  else
    options.set_option("signed-overflow-check", false);

  // check overflow/underflow
  if(cmdline.isset("unsigned-overflow-check"))
    options.set_option("unsigned-overflow-check", true);
  else
    options.set_option("unsigned-overflow-check", false);

  // check for NaN (not a number)
  if(cmdline.isset("nan-check"))
    options.set_option("nan-check", true);
  else
    options.set_option("nan-check", false);

  // check pointers
  if(cmdline.isset("pointer-check"))
    options.set_option("pointer-check", true);
  else
    options.set_option("pointer-check", false);

  // check assertions
  if(cmdline.isset("no-assertions"))
    options.set_option("assertions", false);
  else
    options.set_option("assertions", true);

  // use assumptions
  if(cmdline.isset("no-assumptions"))
    options.set_option("assumptions", false);
  else
    options.set_option("assumptions", true);

  // magic error label
  if(cmdline.isset("error-label"))
    options.set_option("error-label", cmdline.getval("error-label"));

  // generate unwinding assertions
  if(cmdline.isset("cover-assertions"))
    options.set_option("unwinding-assertions", false);
  else
    options.set_option("unwinding-assertions",
      !cmdline.isset("no-unwinding-assertions"));

  // generate unwinding assumptions otherwise
  options.set_option("partial-loops",
   cmdline.isset("partial-loops"));

  // remove unused equations
  options.set_option("slice-formula",
       cmdline.isset("slice-formula"));

  // simplify if conditions and branches
  if(cmdline.isset("no-simplify-if"))
    options.set_option("simplify-if", false);
  else
    options.set_option("simplify-if", true);

  if(cmdline.isset("arrays-uf-always"))
    options.set_option("arrays-uf", "always");
  else if(cmdline.isset("arrays-uf-never"))
    options.set_option("arrays-uf", "never");
  else
    options.set_option("arrays-uf", "auto");

  if(cmdline.isset("dimacs"))
    options.set_option("dimacs", true);

  if(cmdline.isset("refine"))
    options.set_option("refine", true);

  if(cmdline.isset("boolector"))
    options.set_option("boolector", true);

  if(cmdline.isset("mathsat"))
    options.set_option("mathsat", true);

  if(cmdline.isset("cvc"))
    options.set_option("cvc", true);

  if(cmdline.isset("smt1"))
    options.set_option("smt1", true);

  if(cmdline.isset("smt2"))
    options.set_option("smt2", true);

  if(cmdline.isset("fpa"))
    options.set_option("fpa", true);

  if(cmdline.isset("yices"))
    options.set_option("yices", true);

  if(cmdline.isset("z3"))
    options.set_option("z3", true);

  if(cmdline.isset("beautify"))
    options.set_option("beautify", true);

  options.set_option("pretty-names", 
                     !cmdline.isset("no-pretty-names"));

  if(cmdline.isset("outfile"))
    options.set_option("outfile", cmdline.getval("outfile"));

  if(cmdline.isset("clingo"))
  { 
    if(!cmdline.isset("chain-properties")) 
      throw std::string("--clingo requires --chain-properties");

    options.set_option("clingo", true);
  }
  else
    options.set_option("clingo", false);

  if(cmdline.isset("config"))
    options.set_option("config", cmdline.getval("config"));
  else 
    options.set_option("config", std::string(CFG_TMPFILE));
     
  if(cmdline.isset("output-config") )
    options.set_option("output-config", true);

  if(cmdline.isset("incremental-check")) 
    options.set_option("incremental-check", 
			cmdline.getval("incremental-check"));

  if(!cmdline.isset("no-chain") && !cmdline.isset("chain-properties"))
  {
    options.set_option("chain", true); //default
    options.set_option("chain-properties", false);
    options.set_option("unwinding-assertions", false);
  }
  else if(!cmdline.isset("chain") && cmdline.isset("no-chain") && !cmdline.isset("chain-properties"))
  {
    options.set_option("chain", false);
    options.set_option("chain-properties", false);
    options.set_option("unwinding-assertions", false);
  }
  else if(!cmdline.isset("chain") && !cmdline.isset("no-chain") && cmdline.isset("chain-properties"))
  {
    options.set_option("chain", false);
    options.set_option("chain-properties", true);
    options.set_option("unwinding-assertions", false);
  }
  else 
    throw std::string("--chain, --no-chain and --chain-properties are exclusive");
}

/*******************************************************************\

Function: chaincover_parseoptionst::doit

  Inputs:

 Outputs:

 Purpose: invoke main modules

\*******************************************************************/

int chaincover_parseoptionst::doit()
{
  if(cmdline.isset("version"))
  {
    std::cout << CHAINCOVER_VERSION << std::endl;
    return 0;
  }

  //
  // unwinding of transition systems
  //

  if(cmdline.isset("module") ||
     cmdline.isset("gen-interface"))

  {
    error("This version of CBMC has no support for "
          " hardware modules. Please use hw-cbmc.");
    return 1;
  }
  
  register_languages();

  //
  // command line options
  //

  optionst options;
  get_command_line_options(options);
  eval_verbosity();

  if(cmdline.isset("preprocess"))
  {
    preprocessing();
    return 0;
  }

  goto_functionst goto_functions;

  if(get_goto_program(options, goto_functions))
    return 6;
    
  label_properties(goto_functions);

  if(cmdline.isset("show-properties"))
  {
    const namespacet ns(symbol_table);
    show_properties(ns, get_ui(), goto_functions);
    return 0;
  }

  if(set_properties(goto_functions))
    return 7;

  bool retval = doit(options,goto_functions);

#if 0
  // let's log some more statistics
  debug() << "Memory consumption:" << messaget::endl;
  memory_info(debug());
  debug() << eom;
#endif

  return retval;
}

/*******************************************************************\

Function: chaincover_parseoptionst::set_properties

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool chaincover_parseoptionst::set_properties(goto_functionst &goto_functions)
{
  try
  {
    if(cmdline.isset("claim"))
      ::set_properties(goto_functions, cmdline.get_values("claim"));  
  }

  catch(const char *e)
  {
    error(e);
    return true;
  }

  catch(const std::string e)
  {
    error(e);
    return true;
  }
  
  catch(int)
  {
    return true;
  }
  
  return false;
}

/*******************************************************************\

Function: chaincover_parseoptionst::get_goto_program

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
  
bool chaincover_parseoptionst::get_goto_program(
  const optionst &options,
  goto_functionst &goto_functions)
{
  if(cmdline.args.size()==0)
  {
    error("Please provide a program to verify");
    return true;
  }

  try
  {
    if(cmdline.args.size()==1 &&
       is_goto_binary(cmdline.args[0]))
    {
      status("Reading GOTO program from file");

      if(read_goto_binary(cmdline.args[0],
           symbol_table, goto_functions, get_message_handler()))
        return true;
        
      config.ansi_c.set_from_symbol_table(symbol_table);

      if(cmdline.isset("show-symbol-table"))
      {
        show_symbol_table();
        return true;
      }
      
      if(symbol_table.symbols.find(ID_main)==symbol_table.symbols.end())
      {
        error("The goto binary has no entry point; please complete linking");
        return true;
      }
    }
    else
    {
      if(parse()) return true;
      if(typecheck()) return true;
      if(final()) return true;

      // we no longer need any parse trees or language files
      clear_parse();

      if(cmdline.isset("show-symbol-table"))
      {
        show_symbol_table();
        return true;
      }

      if(symbol_table.symbols.find(ID_main)==symbol_table.symbols.end())
      {
        error("No entry point; please provide a main function");
        return true;
      }

      status("Generating GOTO Program");

      goto_convert(symbol_table, goto_functions, ui_message_handler);
    }

    // finally add the library
    status("Adding CPROVER library");      
    link_to_library(symbol_table, goto_functions, ui_message_handler);

    if(process_goto_program(options, goto_functions))
      return true;
  }

  catch(const char *e)
  {
    error(e);
    return true;
  }

  catch(const std::string e)
  {
    error(e);
    return true;
  }
  
  catch(int)
  {
    return true;
  }
  
  catch(std::bad_alloc)
  {
    error("Out of memory");
    return true;
  }
  
  return false;
}

/*******************************************************************\

Function: chaincover_parseoptionst::preprocessing

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
  
void chaincover_parseoptionst::preprocessing()
{
  try
  {
    if(cmdline.args.size()!=1)
    {
      error("Please provide one program to preprocess");
      return;
    }

    std::string filename=cmdline.args[0];

    std::ifstream infile(filename.c_str());

    if(!infile)
    {
      error("failed to open input file");
      return;
    }

    languaget *ptr=get_language_from_filename(filename);

    if(ptr==NULL)
    {
      error("failed to figure out type of file");
      return;
    }

    std::auto_ptr<languaget> language(ptr);
  
    if(language->preprocess(infile, filename, std::cout))
      error("PREPROCESSING ERROR");
  }

  catch(const char *e)
  {
    error(e);
  }

  catch(const std::string e)
  {
    error(e);
  }
  
  catch(int)
  {
  }

  catch(std::bad_alloc)
  {
    error("Out of memory");
  }
}

/*******************************************************************\

Function: chaincover_parseoptionst::process_goto_program

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
  
bool chaincover_parseoptionst::process_goto_program(
  const optionst &options,
  goto_functionst &goto_functions)
{
  try
  {
    namespacet ns(symbol_table);

    if(cmdline.isset("string-abstraction"))
      string_instrumentation(
        symbol_table, get_message_handler(), goto_functions);

    status("Function Pointer Removal");
    remove_function_pointers(symbol_table, goto_functions,
      cmdline.isset("pointer-check"));

    status("Partial Inlining");
    // do partial inlining
    goto_partial_inline(goto_functions, ns, ui_message_handler);
    
    //status("Generic Property Instrumentation");
    // add generic checks
    //   goto_check(ns, options, goto_functions);

  /*  reach_bmc.output_goto_nbac(context.init, context.step, context.input_assume, std::cout);
      return 0;*/
    
    if(cmdline.isset("string-abstraction"))
    {
      status("String Abstraction");
      string_abstraction(symbol_table,
        get_message_handler(), goto_functions);
    }

    // add failed symbols
    // needs to be done before pointer analysis
    add_failed_symbols(symbol_table);
    
    // recalculate numbers, etc.
    goto_functions.update();

    // add loop ids
    goto_functions.compute_loop_numbers();
    
    // show it?
    if(cmdline.isset("show-loops"))
    {
      show_loop_ids(get_ui(), goto_functions);
      return true;
    }

    // show it?
    if(cmdline.isset("show-goto-functions"))
    {
      goto_functions.output(ns, std::cout);
      return true;
    }
  }

  catch(const char *e)
  {
    error(e);
    return true;
  }

  catch(const std::string e)
  {
    error(e);
    return true;
  }
  
  catch(int)
  {
    return true;
  }
  
  catch(std::bad_alloc)
  {
    error("Out of memory");
    return true;
  }
  
  return false;
}



bool chaincover_parseoptionst::doit( 
  const optionst &options,
  goto_functionst &goto_functions)
{
  namespacet ns(symbol_table);

  test_baset* test; 
  config_parser_baset *config_parser = NULL;

  if(options.get_bool_option("chain") || 
     options.get_bool_option("chain-properties")) 
  {
    std::string cfgfile = options.get_option("config");

    if(options.get_bool_option("chain-properties"))
      config_parser = new config_parser_opt_xmlt(cfgfile);
    else
      config_parser = new config_parser_xmlt(cfgfile);

    if((!cmdline.isset("config") && 
        !options.get_bool_option("output-config")) ||
       (cmdline.isset("config") && 
        options.get_bool_option("output-config"))) 
    {
      status("inferring configuration...");
      config_parser->infer(goto_functions,namespacet(symbol_table));
      if(options.get_bool_option("output-config")) 
        status("configuration written to "+cfgfile);
    }
    if(options.get_bool_option("chain-properties"))
    {
      status("Test chains with Hamiltonian path optimisation");
      test = new testchain_optt(options,ui_message_handler,
				(config_parser_opt_xmlt *)config_parser);
    }
    else //chain
    {
      status("Test chains with SAT solver only");
      test = new testchain_satt(options,ui_message_handler,
				(config_parser_xmlt *)config_parser);
    }
  }
  else //no-chain, just cover test goals
  {
    if(/*cmdline.isset("mcdc") ||*/ cmdline.isset("branch")) 
    {
      /*     if(cmdline.isset("mcdc"))
	     mcdc(symbol_table,goto_functions,"");
	     else if(cmdline.isset("branch"))*/
      branchtest(symbol_table,goto_functions);
      add_failed_symbols(symbol_table);
      goto_functions.update();
      goto_functions.compute_loop_numbers();
      label_properties(goto_functions);
      //    goto_functions.output(namespacet(symbol_table),std::cout);
    }
  /*  if(cmdline.isset("accelerate"))
      {
      namespacet ns(symbol_table);
      looppath_select_typet looppath_select_type = LOOPPATH_SELECT_FILTER;
      std::string selectoption = cmdline.getval("looppath-select");
      if(selectoption=="all") looppath_select_type = LOOPPATH_SELECT_ALL;
      if(selectoption=="filter") looppath_select_type = LOOPPATH_SELECT_FILTER;
      if(selectoption=="blocks") looppath_select_type = LOOPPATH_SELECT_BLOCKS;
      if(selectoption=="filter-blocks") looppath_select_type = LOOPPATH_SELECT_FILTER_BLOCKS;

      accelerate_functions(goto_functions, ns,looppath_select_type);
      //    remove_skip(goto_functions);
      goto_functions.update();
      }
  */
    if(cmdline.isset("incremental-check"))
    {
      status("Tests with incremental SAT solver");
      test = new test_sat_incrementalt(options,ui_message_handler);
    }
    else //non-incremental
    {
      status("Tests with non-incremental SAT solver");
      test = new test_sat_nonincrementalt(options,ui_message_handler);
    }
  }

  //compute tests
  test->compute(goto_functions,symbol_table);    
  switch(get_ui())
  {
    case ui_message_handlert::PLAIN:
    {
      if(!test->testvectors.empty())
      {
	if(options.get_bool_option("minimize-test-suite"))
	  test->minimize_test_suite();
        std::ostringstream out;
        test->output_testvectors(out,ns); 
        result(out.str());
        if(cmdline.isset("output-tests"))
	{
	  std::string file = cmdline.getval("output-tests");
          test->output_tests_to_header_file(file,goto_functions,ns);
          status() << "Tests written to header file '" << file << "'" << eom;
	}
      }
      else error() << "No test vectors or chains found." << eom;
      break;
    }
    case ui_message_handlert::XML_UI:
      if(options.get_bool_option("minimize-test-suite"))
        test->minimize_test_suite();
      test->xmloutput_testvectors(std::cout,ns);
      break;
    default: assert(false);
  }

  statistics(test->statistics_to_string());
  bool retval = false;
  if(test->testvectors.empty()) retval = true;

  if(config_parser!=NULL) delete config_parser;
  delete test;

  if(!cmdline.isset("config") && 
     !options.get_bool_option("output-config")) 
    assert(!system(RM_CMD " " CFG_TMPFILE)); //clean up
  
  return retval;
}




/*******************************************************************\

Function: chaincover_parseoptionst::help

  Inputs:

 Outputs:

 Purpose: display command line help

\*******************************************************************/

void chaincover_parseoptionst::help()
{
  std::cout <<
    "\n"
    "* *   ChainCover " CHAINCOVER_VERSION "                          * *\n"
    "* *                      Peter Schrammel                   * *\n"
    "* *                                                        * *\n"
    "* *   based on                                             * *\n"
    "* *   CBMC " CBMC_VERSION " - Copyright (C) 2001-2014 ";
    
  std::cout << "(" << (sizeof(void *)*8) << "-bit version)";
    
  std::cout << "   * *\n";
    
  std::cout <<
    "* *              Daniel Kroening, Edmund Clarke             * *\n"
    "* * Carnegie Mellon University, Computer Science Department * *\n"
    "* *                 kroening@kroening.com                   * *\n"
    "* *        Protected in part by U.S. patent 7,225,417       * *\n"
    "\n"
    "Usage:                        Purpose:\n"
    "\n"
    " chaincover [-?][-h][--help]  show help\n"
    " chaincover file.c ...        source file names\n"
    "\n"
    "Frontend options:\n"
    " --config config-file         set configuration file\n"
    " --output-config              write inferred configuration to file\n"
    "                                given in --config\n"
    " --output-tests header-file   output tests to header file\n"
    " -I path                      set include path (C/C++)\n"
    " -D macro                     define preprocessor macro (C/C++)\n"
    " --preprocess                 stop after preprocessing\n"
    " --16, --32, --64             set width of int\n"
    " --LP64, --ILP64, --LLP64,\n"
    "   --ILP32, --LP32            set width of int, long and pointers\n"
    " --little-endian              allow little-endian word-byte conversions\n"
    " --big-endian                 allow big-endian word-byte conversions\n"
    " --unsigned-char              make \"char\" unsigned by default\n"
    " --show-symbol-table          show symbol table\n"
    " --show-goto-functions        show goto program\n"
    " --ppc-macos                  set MACOS/PPC architecture\n"
    #ifdef _WIN32
    " --i386-macos                 set MACOS/I386 architecture\n"
    " --i386-linux                 set Linux/I386 architecture\n"
    " --i386-win32                 set Windows/I386 architecture (default)\n"
    " --winx64                     set Windows/X64 architecture\n"
    " --gcc                        use GCC as preprocessor\n"
    #else
    #ifdef __APPLE__
    " --i386-macos                 set MACOS/I386 architecture (default)\n"
    " --i386-linux                 set Linux/I386 architecture\n"
    " --i386-win32                 set Windows/I386 architecture\n"
    " --winx64                     set Windows/X64 architecture\n"
    #else
    " --i386-macos                 set MACOS/I386 architecture\n"
    " --i386-linux                 set Linux/I386 architecture (default)\n"
    " --i386-win32                 set Windows/I386 architecture\n"
    " --winx64                     set Windows/X64 architecture\n"
    #endif
    #endif
    " --no-arch                    don't set up an architecture\n"
    " --no-library                 disable built-in abstract C library\n"
    " --round-to-nearest           IEEE floating point rounding mode (default)\n"
    " --round-to-plus-inf          IEEE floating point rounding mode\n"
    " --round-to-minus-inf         IEEE floating point rounding mode\n"
    " --round-to-zero              IEEE floating point rounding mode\n"
#if 0
    "\n"
    "Program instrumentation options:\n"
    " --bounds-check               enable array bounds checks\n"
    " --div-by-zero-check          enable division by zero checks\n"
    " --pointer-check              enable pointer checks\n"
    " --signed-overflow-check      enable arithmetic over- and underflow checks\n"
    " --unsigned-overflow-check    enable arithmetic over- and underflow checks\n"
    " --nan-check                  check floating-point for NaN\n"
    " --show-properties                only show properties\n"
    " --show-loops                 show the loops in the program\n"
    " --no-assertions              ignore user assertions\n"
    " --no-assumptions             ignore user assumptions\n"
    " --error-label label          check that label is unreachable\n"
    " --cover-assertions           check which assertions are reachable\n" 
    "\n"
    "BMC options:\n"
    " --function name              set main function name\n"
    " --claim nr                   only check specific claim\n"
    " --program-only               only show program expression\n"
    " --depth nr                   limit search depth\n"
    " --unwind nr                  unwind nr times\n"
    " --unwindset L:B,...          unwind loop L with a bound of B\n"
    "                              (use --show-loops to get the loop IDs)\n"
    " --unwind-max nr              stop incremental check after nr unwindings\n"
    " --show-vcc                   show the verification conditions\n"
    " --slice-formula              remove assignments unrelated to property\n"
    " --no-unwinding-assertions    do not generate unwinding assertions\n"
    " --partial-loops              permit paths with partial loops\n"
    " --no-pretty-names            do not simplify identifiers\n"
#endif 
    "\n"
    "Backend options:\n"
    " --chain                      compute single chain using SAT solver\n"
    "                                (test goals given as assertions)\n"
    "                                (requires --incremental-check)\n"
    " --no-chain                   compute test vectors without trying to chain\n"
    "                                (test goals given as assertions)\n"
    " --do-not-minimize            do not minimize test suite\n"
    " --chain-properties           compute chain(s)\n"
    "                                (test goals given as property functions)\n"
    " --clingo                     use clingo instead of LKH\n"
    " --path-constraints           use path constraints instead of abstraction\n"
    "                                refinement (requires clingo)\n"
    " --no-multi-chain             do not compute multiple chains\n"
    " --incremental-check L        incrementally check loop L\n"
    //   " --output-harness filename    output test harness\n"
#if 0
    " --dimacs                     generate CNF in DIMACS format\n"
    " --beautify                   beautify the counterexample (greedy heuristic)\n"
    " --smt1                       output subgoals in SMT1 syntax (experimental)\n"
    " --smt2                       output subgoals in SMT2 syntax (experimental)\n"
    " --boolector                  use Boolector (experimental)\n"
    " --mathsat                    use MathSAT (experimental)\n"
    " --cvc                        use CVC3 (experimental)\n"
    " --yices                      use Yices (experimental)\n"
    " --z3                         use Z3 (experimental)\n"
    " --refine                     use refinement procedure (experimental)\n"
    " --outfile filename           output formula to given file\n"
    " --arrays-uf-never            never turn arrays into uninterpreted functions\n"
    " --arrays-uf-always           always turn arrays into uninterpreted functions\n" 
#endif
    "\n"
    "Other options:\n"
    " --version                    show version and exit\n"
    " --xml-ui                     use XML-formatted output\n"
    " --xml-interface              stdio-XML interface\n"
    "\n";
}
