/*******************************************************************\

Module: Test chain generation interface

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_CHAIN_TEMPLATE_H
#define CHAINCOVER_CHAIN_TEMPLATE_H

#include <iostream>
#include <util/time_stopping.h>
#include <limits.h>

#include "config_parser_opt_xml.h"
#include "graph_weighted.h"
#include "reach_template.h"


template<class ContextT, class PropT, class ChainT> 
class testchain_templatet 
{
 public:
  typedef graph_weighted_intidst<PropT> grapht;
  typedef typename grapht::edgest edgest;
  typedef std::set<PropT> propertiest;
  typedef reach_templatet<ContextT,PropT,ChainT> reacht;
  typedef hampath_solver_templatet<nodeidt, PropT, edgeidt> hampath_solvert;
  typedef config_parser_opt_xmlt config_parsert;
  typedef std::list<ChainT> chainst;

  typedef typename reacht::prop_edget prop_edget;
  typedef typename reacht::prop_edgest prop_edgest;
  typedef typename reacht::prop_patht prop_patht;
  typedef typename reacht::pathweightst pathweightst;
  typedef typename grapht::patht patht;
  typedef typename grapht::pathst pathst;

  testchain_templatet(reacht* _reach, 
		      hampath_solvert* _hampath_solver, 
		      config_parsert* _config_parser) 
  {
    reach = _reach;
    hampath_solver = _hampath_solver;
    config_parser = _config_parser;
    reach_bound = INT_MAX;
    chain_bound = 0;
    harness_outfile = "";
  }

/******************************************************************************/
/*** configuration */
/******************************************************************************/

  void set_reach(reacht* _reach) { reach = _reach; }
  void set_hampath_solver(hampath_solvert* _hampath_solver) { hampath_solver = _hampath_solver; }
  void set_config_parser(config_parsert* _config_parser) {  config_parser = _config_parser; }

  // set configuration
  void set_properties(propertiest _properties) { properties = _properties; }
  void set_reach_bound(edgeweightt _reach_bound) { reach_bound = _reach_bound; }
  void set_chain_bound(edgeweightt _chain_bound) { chain_bound = _chain_bound; }
  void set_init(PropT _init) { init = _init; }
  void set_final(PropT _final) { final = _final; }
  void set_harness_outfile(std::string _harness_outfile) { harness_outfile = _harness_outfile; }
  void set_use_path_constraints(bool _use_path_constraints) { use_path_constraints = _use_path_constraints; }
  void set_compute_multichain(bool _compute_multichain) { compute_multichain = _compute_multichain; }

  // read configuration
  void read_config() { 
    config_parser->parse();
    init = config_parser->init; 
    final = config_parser->final;
    properties = config_parser->properties;
    reach_bound = config_parser->reach_bound;
    chain_bound = config_parser->chain_bound;
  }

  // print configuration
  std::ostream &output_config(std::ostream& out) {
    out << "configuration:" << std::endl;
    out << "  init = " << init << std::endl;
    out << "  final = " << final << std::endl;
    out << "  reach bound = " << reach_bound << std::endl;
    out << "  chain bound = " << chain_bound << std::endl;
    out << "  properties = { ";
    for(typename propertiest::iterator it=properties.begin(); 
       it!=properties.end(); it++) {
      out << *it << " ";
    }
    out << " }" << std::endl;
    return out;
  }

  // print configuration to string
  std::string config_to_string() {
    std::ostringstream out; 
    output_config(out);
    return out.str();
  }

  /****************************************************************************/
  // computes the test case chain corresponding to the given shortest path
  /****************************************************************************/
  testvectort get_chain(grapht& graph, patht& path) {
    std::ostringstream out;
    out << "Get chain for path: { ";

    prop_patht proppath;
    pathweightst pathweights;
    typename grapht::patht::iterator pnode=(path).begin(); 
    if(pnode!=(path).end()) { 
      proppath.push_back(graph.get_node_label(*pnode));
      out << graph.get_node_label(*pnode) << " ";
      typename patht::iterator node=pnode; 
      typename grapht::edget edge=typename grapht::edget(*(pnode),*(++node)); 
      for(; node!=(path).end();	
            edge=typename grapht::edget(*(++pnode),*(++node))) {
        pathweights.push_back(graph.get_edge_label(edge));
        proppath.push_back(graph.get_node_label(*node));
        out << graph.get_node_label(*node) << " ";
      }
    }
    out << "}";
    status_msg(out.str());

    // chain repair
    testvectort chain;
    prop_patht failed_path = proppath;
    reach->repair_chain(proppath,pathweights,failed_path,reach_bound,chain);
    if(chain.steps.empty()) 
    {
      //return failed path in path
      patht old_path = path;
      path.clear();
      typename prop_patht::iterator it = proppath.begin();
      typename patht::iterator p = old_path.begin();
      for(;*it!=failed_path.front();it++, p++);
      for(;*it!=failed_path.back();it++, p++) { path.push_back(*p); }
      path.push_back(*p);

      std::ostringstream out;
      out << "failed_path: "; graph.output_path(out,path);
      debug_msg(out.str());
    }

    /*    if(harness_outfile!="") reach->output_harness(chain,proppath,pathweights,
                                                  harness_outfile);
    */

    return chain;
  } 

  /****************************************************************************/
  // refine graph and find shortest path
  /****************************************************************************/
  void refine_and_find_path(grapht& graph, patht& path, nodeidt initid, nodeidt finalid, 
                            edgeweightt chain_bound, pathst path_constraints) {
    if(use_path_constraints) {
      graph.shortest_covering_path(path,initid,finalid,chain_bound,path_constraints);
    }
    else { //use abstraction refinement

      graph.split_node_remove_path(path);

      std::ostringstream out;
      out << "refined graph: "; graph.output(out); debug_msg(out.str());

      graph.covering_path_label_groups(path,initid); 

      clear_out(out); 
      out << "covering path: "; graph.output_path(out,path) << std::endl;
      debug_msg(out.str());

      #if 0
      //no real advantage to optimise here
      grapht graph2; graph.copy(graph2);
      graph2.collapse_label_groups_for_path(path);

      clear_out(out); out << "collapsed graph: "; graph2.output(out); debug_msg(out.str());

      path_constraints.clear(); // no path constraints for abstraction refinement
      patht closure_path;
      graph2.closure();
      graph2.shortest_covering_path(closure_path,initid,finalid,chain_bound,path_constraints);
      //std::cout << "covering path: "; graph2.output_path(std::cout,closure_path) << std::endl;
      if(!closure_path.empty())graph2.map_closure_path(graph,closure_path,path);
      #endif
    }
  }

  /****************************************************************************/
  // compute graph partition admitting covering paths (including init and final)
  /****************************************************************************/
  void compute_partition(grapht graph, nodeidt initid, nodeidt finalid, typename grapht::nodepartitiont& part) {
    graph.covering_path_partition(part);
    typename grapht::nodepartitiont newpart;
    for(typename grapht::nodepartitiont::iterator p = part.begin(); p!=part.end(); p++) {       
      typename grapht::nodest newp;
      newp.insert(initid); newp.insert(finalid); newp.insert(p->begin(),p->end());
	newpart.insert(newp);
      }
    part = newpart;
  }

  /****************************************************************************/
  // compute minimal test case chain
  /****************************************************************************/
  chainst compute_chain() {
    std::ostringstream out;

    status_msg("Building k-reach graph...");

    time_in_kreach = time_in_hampath = time_in_getchain = time_periodt(0);
    absolute_timet time_start(current_time());
    grapht graph(hampath_solver);
    nodeidt initid = graph.add_node(init);
    prop_edgest alledges; //edges to be considered
    for(typename propertiest::iterator it=properties.begin(); 
       it!=properties.end(); it++) {
      graph.add_node(*it); 
      alledges.insert(prop_edget(init,*it));
      alledges.insert(prop_edget(*it,final));
    }
    nodeidt finalid = graph.add_node(final);
    for(typename propertiest::iterator it1=properties.begin(); 
       it1!=properties.end(); it1++) {
      for(typename propertiest::iterator it2=properties.begin(); 
         it2!=properties.end(); it2++) {
        if(it1!=it2) alledges.insert(prop_edget(*it1,*it2));
      }
    }
    //build graph
    edgeweightt k=0;
    typename grapht::patht path;
    while(!alledges.empty() && (k<=reach_bound) && grapht::is_path_empty(path)){

      std::ostringstream out2; out2 << "Computing " << k << "-reach edges..."; 
      status_msg(out2.str());

      prop_edgest kedges = reach->get_k_reach_edges(alledges,k);

      out << k << "-reach edges: ";

      for(typename prop_edgest::iterator it=kedges.begin(); 
          it!=kedges.end(); it++) {

	out << "(" << it->first << "," << it->second << ") ";

        typename grapht::edget edge(graph.get_node_id(it->first),
				    graph.get_node_id(it->second));
	if(graph.exists_edge(edge)) 
          graph.update_edge_label(graph.get_edge_id(edge),k); 
	else graph.add_edge(edge,k);
        alledges.erase(*it);
      }
      out << std::endl;

      out << "all remaining edges: ";
      for(typename prop_edgest::iterator it=alledges.begin(); 
          it!=alledges.end(); it++) {
	out << "(" << it->first << "," << it->second << ") ";
      }
      debug_msg(out.str());

      time_in_kreach += (current_time() - time_start);
      time_start =  current_time();  

      status_msg( "Computing covering path...");

      //check existence of path
      graph.closure();
      clear_out(out); out << "closure graph: "; graph.output(out); debug_msg(out.str());

      if(graph.exists_covering_path(initid,finalid)) {
	pathst path_constraints;
        graph.shortest_covering_path(path,initid,finalid,chain_bound,path_constraints);
      }

      clear_out(out);
      out << "shortest path: "; graph.output_path(out,path) << std::endl;
      out << "length = " << graph.compute_path_weight(path) << std::endl;
      debug_msg(out.str());

      time_in_hampath += (current_time() - time_start);
      time_start = current_time();  

      k++;
    }

    if(grapht::is_path_empty(path) && !compute_multichain) 
      throw "No chain found up to reach bound.";

    status_msg("Computing chain...");

    //compute partition
    typename grapht::nodepartitiont partition;
    compute_partition(graph,initid,finalid,partition);

    clear_out(out); out << "partition: "; graph.output_nodepartition(out,partition);
    debug_msg(out.str());

    chainst chains;
    //for each equivalence class
    for(typename grapht::nodepartitiont::iterator np = partition.begin(); 
        np!=partition.end(); np++) {
      grapht gg;
      graph.copy_subgraph(gg,*np); //extract graph partition

      clear_out(out); out << "subgraph: "; gg.output(out) << std::endl; 
      debug_msg(out.str());

      ChainT chain;
      pathst path_constraints;

      if(grapht::is_path_empty(path)) { //find shortest path (in the multi-chain case)
        //gg.covering_path(path);
        gg.shortest_covering_path(path,initid,finalid,chain_bound,path_constraints);

        clear_out(out);
        out << "shortest path: "; gg.output_path(out,path) << std::endl;
        out << "length = " << gg.compute_path_weight(path) << std::endl;
        debug_msg(out.str());
      }

      time_in_hampath += (current_time() - time_start);
      time_start = current_time();
      while(chain.steps.empty()) { //abstraction refinement loop (to ensure completeness)
        chain = get_chain(gg,path); 
        time_in_getchain += (current_time() - time_start);

        time_start = current_time();
        if(chain.steps.empty()) {
          status_msg("Refine...");

          clear_out(out);
          out << "failed path: "; gg.output_path(out,path);
          debug_msg(out.str());

          path_constraints.insert(path);
          refine_and_find_path(gg,path,initid,finalid,chain_bound,path_constraints);

          clear_out(out); 
          out << "found path: "; gg.output_path(out,path);
          debug_msg(out.str());
        }
        time_in_hampath += (current_time() - time_start);
        time_start = current_time();

        if(grapht::is_path_empty(path)) throw "No chain found up to reach bound.";
      }
      typename grapht::nodest gg_nodes; gg.get_nodes(gg_nodes);
      for(typename grapht::nodest::iterator n_it=gg_nodes.begin();
           n_it!=gg_nodes.end(); ++n_it)
        chain.covered.insert(gg.get_node_label(*n_it));
      chains.push_back(chain);
      path.clear(); //for the multi-chain case
    }

    return chains;
  }

/******************************************************************************/
/*** printing */
/******************************************************************************/

  // print given chain
  std::ostream &output_chain(ChainT chain, std::ostream& out) {
    out << std::endl << "test chain length = " <<chain.size() << 
           std::endl << std::endl;
    out << "test chain: {";
    reach->output_chain(chain,out);
    out << "}" << std::endl;
    return out;
  }

  std::ostream &xmloutput_chain(ChainT chain, std::ostream& out) {
    out << "  <testvector>" << std::endl;
    out << "    <length>" << chain.size() << "</length>" << std::endl;
    out << "    <steps>" << std::endl;
    reach->xmloutput_chain(chain,out);
    out << "    </steps>" << std::endl;
    out << "  </testvector>" << std::endl;
    return out;
  }

  // print chain to string
  std::string chain_to_string(ChainT chain) {
    std::ostringstream out; 
    output_chain(chain,out);
    return out.str();
  }

  // prints statistics
  std::ostream &output_statistics(std::ostream& out) {
    out << "statistics:" << std::endl;
    out << "  time spent in building k-reach graph:   ";
    out << time_in_kreach << std::endl;
    out << "  time spent in hamiltonian path solving: ";
    out << time_in_hampath << std::endl;
    out << "  time spent in computing chain:          ";
    out << time_in_getchain << std::endl;
    out << "  total time:                             ";
    out << (time_in_kreach+time_in_hampath+time_in_getchain) << std::endl;

    return out;
  }

  // print statistics to string
  std::string statistics_to_string() {
    std::ostringstream out; 
    output_statistics(out);
    return out.str();
  }

 protected:
  hampath_solvert* hampath_solver;
  reacht* reach;
  config_parsert* config_parser;
  edgeweightt reach_bound; // compute k-reachability information up to this bound
  edgeweightt chain_bound; //bound for solution path being acceptable
  PropT init, final;
  propertiest properties; 
  std::string harness_outfile;
  bool use_path_constraints;
  bool compute_multichain;

  time_periodt time_in_kreach, time_in_hampath, time_in_getchain;

  virtual void debug_msg(std::string msg) = 0;
  virtual void status_msg(std::string msg) = 0;

 private:
  void clear_out(std::ostringstream& out) {
    out.seekp(0); out.str(""); out.clear();      
  }

};

#endif
