/*******************************************************************\

Module: Labelled directed graphs

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_GRAPH_TEMPLATE_H
#define CHAINCOVER_GRAPH_TEMPLATE_H

#include <util/std_types.h>
#include <set>
#include <iostream>


/*******************************************************************/
/** labelled graph template */
/*******************************************************************/
template<class NodeIdT, class NodeLabelT, class EdgeIdT, class EdgeLabelT> class graph_templatet {

public:
  typedef std::pair<NodeIdT,NodeIdT> edget;
  typedef std::list<NodeIdT> patht;
  typedef std::set<patht> pathst;
  typedef std::set<edget> edgest;
  typedef std::set<NodeIdT> nodest;
  typedef std::map<NodeIdT,NodeIdT> nodeidmapt;
 

/*******************************************************************/
/*** adding and retrieving nodes and edges */
/*******************************************************************/

  NodeIdT add_node(NodeLabelT label) {
    NodeIdT id = new_node_id();
    nodelabels.insert(std::pair<NodeIdT,NodeLabelT>(id,label));
    return id;
  }

  void add_node_with_id(NodeIdT id, NodeLabelT label) {
    if(nodelabels.find(id)!=nodelabels.end())  throw "ERROR: duplicated node id";
    nodelabels.insert(std::pair<NodeIdT,NodeLabelT>(id,label));
  }

  NodeLabelT get_node_label(NodeIdT id) {
    return nodelabels.at(id);
  }
  NodeIdT get_node_id(NodeLabelT label) { 
    for(typename std::map<NodeIdT,NodeLabelT>::iterator node=nodelabels.begin();
	node!=nodelabels.end(); ++node) {
      if(node->second==label) return node->first;
    }
    throw "ERROR: node not found";
  }

  nodest get_node_ids(NodeLabelT label) { 
    nodest foundnodes;
    for(typename std::map<NodeIdT,NodeLabelT>::iterator node=nodelabels.begin();
	node!=nodelabels.end(); ++node) {
      if(node->second==label) foundnodes.insert(node->first);
    }
    return foundnodes;
  }

  EdgeIdT add_edge(edget edge, EdgeLabelT label) {
    EdgeIdT id = new_edge_id();
    edges.insert(std::pair<EdgeIdT,edget>(id,edge));
    edgelabels.insert(std::pair<EdgeIdT,EdgeLabelT>(id,label));
    return id;
  }

  void add_edge_with_id(EdgeIdT id, edget edge, EdgeLabelT label) {
    if(edgelabels.find(id)!=edgelabels.end())  throw "ERROR: duplicated edge id";
    edges.insert(std::pair<EdgeIdT,edget>(id,edge));
    edgelabels.insert(std::pair<EdgeIdT,EdgeLabelT>(id,label));
  }

  edget get_edge(EdgeIdT id) {
    return edges.at(id);
  }

  EdgeLabelT get_edge_label(EdgeIdT id) {
    return edgelabels.at(id);
  }

  //returns label of first edge found
  EdgeLabelT get_edge_label(edget e) { 
    for(typename std::map<EdgeIdT,edget>::iterator edge=edges.begin();
	edge!=edges.end(); ++edge) {
      if(edge->second==e) return edgelabels.at(edge->first);
    }
    throw "ERROR: edge not found";
  }

  void update_edge_label(EdgeIdT id, EdgeLabelT label) {
    edgelabels.at(id)=label;
  }
  
  //returns first edge found
  EdgeIdT get_edge_id(edget e) { 
    for(typename std::map<EdgeIdT,edget>::iterator edge=edges.begin();
	edge!=edges.end(); ++edge) {
      if(edge->second==e) return edge->first;
    }
    throw "ERROR: edge not found";
  }

  void remove_edge(EdgeIdT id) {
    edges.erase(id);
    edgelabels.erase(id);
  }

  void remove_node(NodeIdT id) {
    nodelabels.erase(id);
    std::map<EdgeIdT,edget> newedges;
    std::map<EdgeIdT,EdgeLabelT> newedgelabels;
    for(typename std::map<EdgeIdT,edget>::iterator edge=edges.begin();
	edge!=edges.end(); ++edge) {
      if(edge->second.first!=id && edge->second.second!=id) {
        newedges.insert(*edge);
        newedgelabels.insert(
		      std::pair<EdgeIdT,EdgeLabelT>(edge->first,get_edge_label(edge->first)));
      }
    }
    edges.clear();
    edges.insert(newedges.begin(),newedges.end());
    edgelabels.clear();
    edgelabels.insert(newedgelabels.begin(),newedgelabels.end());
  }

  bool exists_edge(edget e) {
    for(typename std::map<EdgeIdT,edget>::iterator edge=edges.begin();
	edge!=edges.end(); ++edge) {
      if(edge->second==e) return true;
    }
    return false;
  }

  int get_node_count() { return nodelabels.size(); }
  int get_edge_count() { return edges.size(); }

  edgest outgoing_edges(NodeIdT n) { 
    edgest ee;
    for(typename std::map<EdgeIdT,edget>::iterator it = edges.begin(); 
        it!=edges.end(); it++) {
      if(it->second.first==n) {
        ee.insert(it->second);
      }
    }
    return ee;
  }

  edgest incoming_edges(NodeIdT n) { 
    edgest ee;
    for(typename std::map<EdgeIdT,edget>::iterator it = edges.begin(); 
        it!=edges.end(); it++) {
      if(it->second.second==n) {
        ee.insert(it->second);
      }
    }
    return ee;
  }

  void get_edges(edgest& _edges) { 
    for(typename std::map<EdgeIdT,edget>::iterator edge=edges.begin();
	edge!=edges.end(); ++edge) {
      _edges.insert(edge->second);
    }
  }

  void get_nodes(nodest& _nodes) { 
    for(typename std::map<NodeIdT,NodeLabelT>::iterator node=nodelabels.begin();
	node!=nodelabels.end(); ++node) {
      _nodes.insert(node->first);
    }
  }

/*******************************************************************/
/*** printing */
/*******************************************************************/

  std::ostream &output(std::ostream& out) {
    out << "nodes: " << std::endl;
    for(typename std::map<NodeIdT,NodeLabelT>::iterator node=nodelabels.begin();
	node!=nodelabels.end(); ++node) {
      out << node->first << "(" << node->second << ")" << std::endl;
    }
    out << "edges: " << std::endl;
    for(typename std::map<EdgeIdT,EdgeLabelT>::iterator e=edgelabels.begin();
	e!=edgelabels.end(); ++e) {
      edget edge = edges.at(e->first);
      out << e->first << "(" << edge.first << "," << edge.second << ")" 
          << "(" << e->second << ")" << std::endl;
    }
    out << "------" << std::endl;
    return out;
  }

  std::ostream &output_path(std::ostream& out, patht& path) {
    out << "{";
  for(typename patht::iterator node=(path).begin(); 
      node!=(path).end(); ++node) {
      out << *node << " "; 
    }
    out << "}";
    return out;
  }

  static bool is_path_empty(patht& path) {
    return path.empty();
  }

/*******************************************************************/
/*** copying */
/*******************************************************************/

  virtual void copy(graph_templatet<NodeIdT,NodeLabelT,EdgeIdT,EdgeLabelT>& cg) {
    for(typename std::map<NodeIdT,NodeLabelT>::iterator n = nodelabels.begin(); 
          n != nodelabels.end(); ++n) {
      cg.add_node_with_id(n->first,n->second);
    }
    for(typename std::map<EdgeIdT,edget>::iterator e = edges.begin(); 
          e != edges.end(); ++e) {
      cg.add_edge_with_id(e->first,e->second,get_edge_label(e->first));
    }
  }

  virtual void copy_subgraph(
      graph_templatet<NodeIdT,NodeLabelT,EdgeIdT,EdgeLabelT>& cg, nodest nodes) {
    for(typename std::map<NodeIdT,NodeLabelT>::iterator n = nodelabels.begin(); 
          n != nodelabels.end(); ++n) {
      if(nodes.find(n->first)!=nodes.end()) cg.add_node_with_id(n->first,n->second);
    }
    for(typename std::map<EdgeIdT,edget>::iterator e = edges.begin(); 
          e != edges.end(); ++e) {
      if(nodes.find(e->second.first)!=nodes.end() && 
         nodes.find(e->second.second)!=nodes.end()) {
        cg.add_edge_with_id(e->first,e->second,get_edge_label(e->first));
      }
    }
  }

  virtual void copy_map_nodeids(graph_templatet<NodeIdT,NodeLabelT,EdgeIdT,EdgeLabelT>& cg, nodeidmapt m) {
    for(typename std::map<NodeIdT,NodeLabelT>::iterator n = nodelabels.begin(); 
          n != nodelabels.end(); ++n) {
      cg.add_node_with_id(m.at(n->first),n->second);
    }
    for(typename std::map<EdgeIdT,edget>::iterator e = edges.begin(); 
          e != edges.end(); ++e) {
      cg.add_edge_with_id(e->first,edget(m.at(e->second.first),m.at(e->second.second)),
        get_edge_label(e->first));
    }
  }

  static void reverse_nodeidmap(nodeidmapt& m) {
    nodeidmapt newm;
    for(typename nodeidmapt::iterator n = m.begin(); n != m.end(); ++n) {  
      assert(newm.find(n->second)==newm.end()); //m must be injective
      newm.insert(std::pair<NodeIdT,NodeIdT>(n->second,n->first));
    }  
    m = newm;
  }

  static void path_map_nodeids(patht& path, nodeidmapt m) {
    patht newp;
    for(typename patht::iterator node=(path).begin(); node!=(path).end(); ++node) {
      newp.push_back(m.at(*node));
    }
    path = newp;
  }

protected:
  std::map<EdgeIdT,edget> edges;
  std::map<NodeIdT,NodeLabelT> nodelabels;
  std::map<EdgeIdT,EdgeLabelT> edgelabels;

  virtual EdgeIdT new_edge_id() = 0;  
  virtual NodeIdT new_node_id() = 0; 
 
};

#endif
