/*******************************************************************\

Module: Weighted directed graphs

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_GRAPH_WEIGHTED_H
#define CHAINCOVER_GRAPH_WEIGHTED_H

#include "graph_template.h"
#include <iostream>

typedef unsigned edgeweightt;

template<class NodeIdT, class PropT, class EdgeIdT> class graph_weightedt;

/** interface for hamiltonian path solver */
template<class NodeIdT, class PropT, class EdgeIdT>
class hampath_solver_templatet {
  public:
  typedef graph_weightedt<NodeIdT, PropT, EdgeIdT> grapht;
  virtual ~hampath_solver_templatet() {}
  virtual void solve(typename grapht::patht& path, grapht* graph, NodeIdT init, NodeIdT final, edgeweightt bound, typename grapht::pathst path_constraints){}
};

//Weighted graph template
template<class NodeIdT, class PropT, class EdgeIdT>
class graph_weightedt : public graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt> {

public: 
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::edget edget;
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::patht patht;
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::edgest edgest;
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::pathst pathst;
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodest nodest;
  typedef std::set<nodest> nodepartitiont;

  typedef std::map<NodeIdT,PropT> nodelabelst;
  typedef std::map<EdgeIdT,edgeweightt> edgelabelst;
  typedef typename graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodeidmapt nodeidmapt;
  typedef std::map<EdgeIdT,NodeIdT> closure_mapt;
  closure_mapt closure_edges;
 
  graph_weightedt() {}

  graph_weightedt(hampath_solver_templatet<NodeIdT, PropT, EdgeIdT>* solver) {
    hampath_solver = solver;
  }

  void set_hampath_solver(hampath_solver_templatet<NodeIdT, PropT, EdgeIdT>* solver) {
    hampath_solver = solver;
  }

  /******************************************************************************/
  /** returns a path shorter than bound or the shortest path from init to final that passes through all nodes */
  /******************************************************************************/
  void shortest_covering_path(patht& path, NodeIdT init, NodeIdT final, edgeweightt bound, 
      pathst path_constraints) {
    path.clear();
    hampath_solver->solve(path,this,init,final,bound,path_constraints);
  }

  /******************************************************************************/
  // compute sum of edge weights over the given path
  /******************************************************************************/
  edgeweightt compute_path_weight(patht& path) {
    edgeweightt w = 0;
    typename patht::iterator pnode=(path).begin(); 
    if(pnode!=(path).end()) { 
      typename patht::iterator node=pnode; 
      edget edge=edget(*(pnode),*(++node)); 
      for(; node!=(path).end();	
	  edge=edget(*(++pnode),*(++node))) {
        w += this->get_edge_label(edge);
      }
    }
    return w;
  }

  /******************************************************************************/
  // returns the maximum edge weight
  /******************************************************************************/
  edgeweightt get_max_weight() {
    edgeweightt w = 0;
    for(typename std::map<EdgeIdT,edgeweightt>::iterator it = 
        graph_templatet<NodeIdT, PropT, EdgeIdT, 
                        edgeweightt>::edgelabels.begin(); 
        it != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                              edgeweightt>::edgelabels.end(); it++) {
      if(it->second>w) w=it->second;
    }
    return w;
  }

  /******************************************************************************/
  //adds transitive closure edges (without self-loops) using Floyd-Warshall
  /******************************************************************************/
  void closure() {
    for(typename nodelabelst::iterator n1 = 
        graph_templatet<NodeIdT, PropT, EdgeIdT, 
                        edgeweightt>::nodelabels.begin();
        n1 != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                              edgeweightt>::nodelabels.end(); 
        ++n1) {
      for(typename nodelabelst::iterator n2 = 
          graph_templatet<NodeIdT, PropT,EdgeIdT, 
                          edgeweightt>::nodelabels.begin();
          n2 != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                            edgeweightt>::nodelabels.end(); 
          ++n2) {
        for(typename nodelabelst::iterator n3 = 
            graph_templatet<NodeIdT, PropT, EdgeIdT, 
                            edgeweightt>::nodelabels.begin(); 
            n3 != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                                  edgeweightt>::nodelabels.end(); ++n3) {
  	  edget e1(n2->first,n3->first);
	  edget e2(n2->first,n1->first);
	  edget e3(n1->first,n3->first);
          bool exists1 = this->exists_edge(e1);
          bool exists23 = this->exists_edge(e2) && this->exists_edge(e3);
          if(exists23 && !exists1) {
            edgeweightt w2 = this->get_edge_label(e2);
            edgeweightt w3 = this->get_edge_label(e3);
	    if(n2->first!=n3->first) { //no transitive self-loops
	      //  std::cout << "adding " << n2->first << " -" << w2 << "-> " << n1->first << " -" << w3 << "-> " << n3->first << " with weight " << (w2+w3) << std::endl;
              EdgeIdT id = this->add_edge(e1,w2+w3); 
              closure_edges.insert(typename closure_mapt::value_type(id,n1->first));
	    }
  	  }
          if(exists23 && exists1) {
            edgeweightt w1 = this->get_edge_label(e1);
            edgeweightt w2 = this->get_edge_label(e2);
            edgeweightt w3 = this->get_edge_label(e3);
            if(w1>w2+w3) {
	      //	      std::cout << "updating " << n2->first << " -" << w2 << "-> " << n1->first << " -" << w3 << "-> " << n3->first << " with weight " << (w2+w3) << std::endl;
	      this->update_edge_label(this->get_edge_id(e1),w2+w3);
	    }
	  }
        }
      }
    }
  }

  /******************************************************************************/
  //map path in closure graph
  /******************************************************************************/
  void map_recursive(graph_weightedt<NodeIdT,PropT,EdgeIdT>& ncg, NodeIdT n1,NodeIdT n2, patht& ncpath) {
  #if 0
    std::cout << "map_recursive (" << n1 << "," << n2 << ")" << std::endl;
    std::cout << "map path: "; output_path(std::cout,ncpath) << std::endl;
  #endif
    edget e1(n1,n2);
    if(!ncg.exists_edge(e1)) {
  #if 0
      std::cout << "do map" << std::endl;
  #endif
      NodeIdT n = closure_edges[get_edge_id(e1)];
      map_recursive(ncg,n1,n,ncpath);
      map_recursive(ncg,n,n2,ncpath);
    }
    else ncpath.push_back(n2);
  }

  void map_closure_path(graph_weightedt<NodeIdT,PropT,EdgeIdT>& ncg, patht path, patht& ncpath) {
    ncpath.clear();
    typename patht::iterator lastn = path.begin();
    ncpath.push_back(*lastn);
    typename patht::iterator n = path.begin(); n++;
    for(; n!=path.end(); lastn = n++) {
      map_recursive(ncg,*lastn,*n,ncpath);
    }
  }

  /******************************************************************************/
  //checks the existence of a path from start to end 
  //  assumes that the graph is transitively closed
  /******************************************************************************/
  bool exists_covering_path(NodeIdT start, NodeIdT end) {
    for(typename nodelabelst::iterator n1 = 
        graph_templatet<NodeIdT, PropT, EdgeIdT, 
                        edgeweightt>::nodelabels.begin(); 
        n1 != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                              edgeweightt>::nodelabels.end(); ++n1) {
      if(n1->first != start) {
	edget e(start,n1->first);
        if(!this->exists_edge(e)) {
	  /*	  std::cout << "necessary edge not found: " << 
		  start << " -> " << n1->first << std::endl; */
	  return false;
	}
      }
      if(n1->first != end) {
	  edget e(n1->first,end);
        if(!this->exists_edge(e)) {
	  /*	  std::cout << "necessary edge not found: " << 
		  n1->first << " -> " << end << std::endl;*/
	  return false;
	}
      }
      for(typename nodelabelst::iterator n2 = 
          graph_templatet<NodeIdT, PropT, EdgeIdT, 
                          edgeweightt>::nodelabels.begin(); 
          n2 != graph_templatet<NodeIdT, PropT, EdgeIdT, 
                                edgeweightt>::nodelabels.end(); ++n2) {
        if(n1->first != n2->first && 
             n1->first != start && n1->first != end &&
             n2->first != start && n2->first != end) {
  	  edget e1(n1->first,n2->first);
  	  edget e2(n2->first,n1->first);
          if(!this->exists_edge(e1) && !this->exists_edge(e2)) {
	    /*	    std::cout << "necessary edge not found: " << 
		    n1->first << " <-|-> " << n2->first << std::endl; */
  	    return false;
	  }
	}
      }
    }
    return true;
  }
    
  /******************************************************************************/
  //computes a covering path assuming that the graph is transitively closed
  /******************************************************************************/
  void covering_path(patht& p) {
    p.clear();
    NodeIdT last;
    for(typename nodelabelst::iterator n = graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
          n != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n) {
      if(!p.empty()) {
	edget e(last,n->first);
        if(this->exists_edge(e)) { //add after last element?
          p.push_back(n->first);
          last = n->first;
	}
        else { //not at end, so try after previous positions
          typename patht::iterator v = p.end();
          while(v != p.begin()) {
	    --v;
   	    edget e(*v,n->first);
            if(this->exists_edge(e)) { 
              p.insert(++v,n->first);
              break;
	    }
	  }
          if(v == p.begin()) { //try before first position
   	    edget e(n->first,*v);
            if(this->exists_edge(e)) { 
              p.push_front(n->first);
	    }
            else {  //no covering path
	      p.clear();
              return;
	    }
	  }
	}
      }
      else { //path empty, so just insert
        p.push_back(n->first);
        last = n->first;
      }
    }
  }

  /******************************************************************************/
  //computes a path covering all node labels (only one node of a group of nodes with identical labels is covered) (depth-first with backtracking)
  /******************************************************************************/
  bool check_covered(patht p) {
    std::set<PropT> labels;
    for(typename nodelabelst::iterator n = 
     graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
     n != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n) {
      labels.insert(n->second);
    }
    for(typename patht::iterator n = p.begin(); n != p.end(); ++n) {
      labels.erase(this->get_node_label(*n));
    }
    return labels.empty();
  }

  bool path_contains(patht p, NodeIdT n) {
    for(typename patht::iterator it = p.begin(); it!=p.end(); it++) {
      if(*it==n) return true;
    }
    return false;
  }

  bool covering_path_label_groups_recursive(NodeIdT n, patht& p,nodest avoidn, edgest avoide) {    

  #if 0
    std::cout << "cover_recursive (" << n << ")" << std::endl;
    std::cout << "path: "; output_path(std::cout,p) << std::endl;
  #endif
    p.push_back(n);
    if(check_covered(p)) return true;
    nodest thisavoidn = avoidn;
    nodest labelgroupn = this->get_node_ids(this->get_node_label(n));
    labelgroupn.erase(n);
    thisavoidn.insert(labelgroupn.begin(),labelgroupn.end());
    edgest thisavoide = avoide;
    edgest tovisitede;
    edgest noute = this->outgoing_edges(n);
    for(typename edgest::iterator it = noute.begin(); it!=noute.end(); it++) {
      if(avoide.find(*it)!=avoide.end()) continue;
      if(avoidn.find(it->second)!=avoidn.end()) continue;
      if(path_contains(p,it->second)) {
        tovisitede.insert(*it);
        continue;
      }
      thisavoide.insert(*it);
      if(covering_path_label_groups_recursive(it->second,p,thisavoidn,
					      thisavoide)) return true;
    }
    for(typename edgest::iterator it = tovisitede.begin(); 
          it!=tovisitede.end(); it++) {
      thisavoide.insert(*it);
      if(covering_path_label_groups_recursive(it->second,p,thisavoidn,
					      thisavoide)) return true;
    }
    p.pop_back();
    return false;
  }

  void covering_path_label_groups(patht& p, NodeIdT start_node) { 
    p.clear();
    covering_path_label_groups_recursive(start_node,p,nodest(),edgest());
  }

  /******************************************************************************/
  //computes a partition of nodes such that each admits a covering path
  /******************************************************************************/
  void covering_path_partition(nodepartitiont& part) {
    //compute edges that violate the covering path condition
    std::set<edget> violating_edges;
    for(typename nodelabelst::iterator n1 = 
          graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
          n1 != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n1) {
      for(typename nodelabelst::iterator n2 = 
          graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
          n2 != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n2) {
	if(n1->first!=n2->first) {
          edget e1(n1->first,n2->first);
          edget e2(n2->first,n1->first);
          if(!this->exists_edge(e1) && !this->exists_edge(e2) && 
             violating_edges.find(e1)==violating_edges.end() &&
             violating_edges.find(e2)==violating_edges.end()) {
            violating_edges.insert(e1);
	  }
	}
      }
    }
    //build positive/negative equivalence classes
    typedef std::pair<nodest,nodest> nodespairt;
    std::set<nodespairt> pnpart;
    nodest cnodes;
    for(typename std::set<edget>::iterator e=violating_edges.begin(); 
        e!=violating_edges.end(); e++) {
      NodeIdT n1 = e->first;
      NodeIdT n2 = e->second;
      cnodes.insert(n1); cnodes.insert(n2); //remember processed nodes
      if(pnpart.empty()) { //no equivalence classes yet
        nodest ps1,ns1,ps2,ns2;
        ps1.insert(n1); ns1.insert(n2);
        ps2.insert(n2); ns2.insert(n1);
        pnpart.insert(nodespairt(ps1,ns1));
        pnpart.insert(nodespairt(ps2,ns2));
      }
      else { // iterate through all equivalence classes
        std::set<nodespairt> newpnpart;
        for(typename std::set<nodespairt>::iterator np=pnpart.begin(); np!=pnpart.end(); np++) {
          nodest p = np->first;
          nodest n = np->second;
          if(p.find(n1)!=p.end()) { 
            if(p.find(n2)!=p.end()) { //conflict -> split class
              nodest ps(p), ns(n);
              p.erase(n1); n.insert(n1);
              ps.erase(n2); ns.insert(n2);
              newpnpart.insert(nodespairt(ps,ns));
	    }
            else { //add negative element
              n.insert(n2);
	    }
	  }
          else { //n1 not in p
            if(p.find(n2)!=p.end()) { //add negative element
              n.insert(n1);
	    }
            else { //no conflict
              if(n.find(n1)!=n.end()) { //add positive element
                 p.insert(n2);
	      }
              else if(n.find(n2)!=n.end()) { //add positive element
                 p.insert(n1);
	      }
              else { // duplicate class
                nodest ps(p), ns(n);
                p.insert(n1); n.insert(n2);
                ps.insert(n2); ns.insert(n1);
                newpnpart.insert(nodespairt(ps,ns));
              }
	    }
	  }
          newpnpart.insert(nodespairt(p,n));
        }
        pnpart = newpnpart;
      }
    }
    //extract positive classes
    part.clear();
    for(typename std::set<nodespairt>::iterator np=pnpart.begin(); np!=pnpart.end(); np++) {
      if(!np->first.empty()) part.insert(np->first);
    }
    if(!part.empty()) { //add non-processed nodes to first class
      nodest p = *part.begin();
      for(typename nodelabelst::iterator n = 
          graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
          n != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n) {
        if(cnodes.find(n->first)==cnodes.end()) p.insert(n->first);
      }
      part.erase(part.begin());
      part.insert(p);
    }
    else { //return whole node set as single class
      nodest p;
      for(typename nodelabelst::iterator n = 
          graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.begin(); 
          n != graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::nodelabels.end(); ++n) {
        p.insert(n->first);
      }
      part.insert(p);
    }
  }

/******************************************************************************/
/*** printing */
/******************************************************************************/

  // prints a nodes partition
  std::ostream &output_nodepartition(std::ostream& out, nodepartitiont& np) {
    out << "{ ";
    for(typename nodepartitiont::iterator nn=np.begin(); nn!=np.end(); ++nn) {
      out << "{ ";
      for(typename nodest::iterator n=nn->begin(); n!=nn->end(); ++n) {
        out << *n << " "; 
      }
      out << "} ";
    }
    out << "}";
    return out;
  }

/******************************************************************************/
/*** functions for abstraction refinement */
/******************************************************************************/

  //splits a node to remove a path
  void split_node_remove_path(patht path) {
    assert(path.size()==3);
    typename patht::iterator p = path.begin();
    NodeIdT s = *p;
    NodeIdT n = *(++p);
    NodeIdT t = *(++p);
    nodest nns = this->get_node_ids(this->get_node_label(n));
    for(typename nodest::iterator nn=nns.begin();nn!=nns.end();nn++) {
      edget sn = edget(s,*nn);
      if(!this->exists_edge(sn)) continue;
      NodeIdT n1 = this->add_node(this->get_node_label(n));    
      // copy edges(s,n) -> (s,n1)
      edgest nine = this->incoming_edges(*nn);
      for(typename edgest::iterator it = nine.begin(); 
	  it!=nine.end(); it++) {
        edgeweightt nntw = this->get_edge_label(*it);
        this->add_edge(edget(it->first,n1),nntw);
      }
      // remove edge (s,n)
      EdgeIdT snid = this->get_edge_id(sn);
      this->remove_edge(snid);
      // copy edges(n,-t) -> (n1,-t)
      edgest noute = this->outgoing_edges(*nn);
      for(typename edgest::iterator it = noute.begin(); 
	  it!=noute.end(); it++) {
	if(it->second!=t) {
	  edgeweightt nntw = this->get_edge_label(*it);
	  this->add_edge(edget(n1,it->second),nntw);
	}
      }
    }
  }

  // removes nodes with identical labels that are not in path
  void collapse_label_groups_for_path(patht path) {
    for(typename patht::iterator p = path.begin();p!=path.end();p++) {
      nodest nn = this->get_node_ids(get_node_label(*p));
      for(typename nodest::iterator n = nn.begin();n!=nn.end();n++) {
        if(*n!=*p) {  //for each label group node not in path
          //for the cross product of incoming,outgoing edges
          edgest eins = this->incoming_edges(*n);
          edgest eouts = this->outgoing_edges(*n);
          //add or update (if weight is lower) edges
          for(typename edgest::iterator ein = eins.begin();ein!=eins.end();ein++) {
            for(typename edgest::iterator eout = 
		  eouts.begin();eout!=eouts.end();eout++) {
              edget e = edget(ein->first,eout->second);
              edgeweightt w = this->get_edge_label(*ein)+
		this->get_edge_label(*eout);
              if(exists_edge(e)) {
                EdgeIdT eid = this->get_edge_id(e);
                edgeweightt wold = this->get_edge_label(eid);
                if(wold>w) this->update_edge_label(eid,w);
	      }
              else this->add_edge(e,w);
  	    }
	  }
          this->remove_node(*n); //remove node and edges
	}
      }
    }
  }

/******************************************************************************/
/*** copying */
/******************************************************************************/

  virtual void copy(graph_weightedt<NodeIdT,PropT,EdgeIdT>& cg) {
    cg.set_hampath_solver(hampath_solver);
    graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::copy(cg);
  }

  virtual void copy_map_nodeids(graph_weightedt<NodeIdT,PropT,EdgeIdT>& cg, nodeidmapt m) {
    cg.set_hampath_solver(hampath_solver);
    graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::copy_map_nodeids(cg,m);
  }

  virtual void copy_subgraph(graph_weightedt<NodeIdT,PropT,EdgeIdT>& cg, nodest nodes) {
    cg.set_hampath_solver(hampath_solver);
    graph_templatet<NodeIdT, PropT, EdgeIdT, edgeweightt>::copy_subgraph(cg,nodes);
  }

protected:  
  hampath_solver_templatet<NodeIdT, PropT, EdgeIdT>* hampath_solver;
  virtual EdgeIdT new_edge_id() = 0;  
  virtual NodeIdT new_node_id() = 0; 
};

/************ partial instantiation with integer node and edge ids ************/

typedef int nodeidt;
typedef int edgeidt;

//weighted graph template with integer ids
template<class PropT>
class graph_weighted_intidst : public graph_weightedt<nodeidt, PropT, edgeidt> {
public:
  graph_weighted_intidst() : nextnodeid(0), nextedgeid(0) {}

  graph_weighted_intidst(hampath_solver_templatet<nodeidt, PropT, edgeidt>* solver) : graph_weightedt<nodeidt, PropT, edgeidt>(solver), nextnodeid(0), nextedgeid(0) {
  }

  virtual void copy(graph_weighted_intidst<PropT>& cg) {
    graph_weightedt<nodeidt, PropT, edgeidt>::copy(cg);
    cg.update_nextids();
  }

  virtual void copy_map_nodeids(graph_weighted_intidst<PropT>& cg, 
      typename graph_weightedt<nodeidt, PropT, edgeidt>::nodeidmapt m) {
    graph_weightedt<nodeidt, PropT, edgeidt>::copy_map_nodeids(cg,m);
    cg.update_nextids();
  }

  virtual void copy_subgraph(graph_weighted_intidst<PropT>& cg, 
      typename graph_weighted_intidst<PropT>::nodest nodes) {
    graph_weightedt<nodeidt, PropT, edgeidt>::copy_subgraph(cg,nodes);
    cg.update_nextids();
  }

  //update next node and edge ids
  void update_nextids() {
    for(typename graph_weightedt<nodeidt, PropT, edgeidt>::nodelabelst::iterator n = 
          graph_templatet<nodeidt, PropT, edgeidt, edgeweightt>::nodelabels.begin(); 
          n != graph_templatet<nodeidt, PropT, edgeidt, edgeweightt>::nodelabels.end(); ++n) {
      if(n->first>=nextnodeid) nextnodeid = n->first+1;
    }
    for(typename graph_weightedt<nodeidt, PropT, edgeidt>::edgelabelst::iterator e = 
          graph_templatet<nodeidt, PropT, edgeidt, edgeweightt>::edgelabels.begin(); 
          e != graph_templatet<nodeidt, PropT, edgeidt, edgeweightt>::edgelabels.end(); ++e) {
      if(e->first>=nextedgeid) nextedgeid = e->first+1;
    }  
  }

private:
  edgeidt nextnodeid; 
  nodeidt nextedgeid;

protected:
  virtual edgeidt new_edge_id() { return nextedgeid++; }
  virtual nodeidt new_node_id() { return nextnodeid++; }

};

#endif
