/*******************************************************************\

Module: Clingo interface

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_HAMPATH_SOLVER_CLINGO_H
#define CHAINCOVER_HAMPATH_SOLVER_CLINGO_H

#include "graph_weighted.h"
#include "clingo_parser.h"
#include <util/message.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>

#define CLINGO_CMD "clingo -n 0"
#define OUTFILE "__out__.tmp"
#define INFILE "__in__.tmp"
#define RM_CMD "rm -f"

#define DEBUG_LEVEL (get_message_handler().get_verbosity()>=9)

template<class PropT>
class hampath_solver_clingot : public hampath_solver_templatet<nodeidt, PropT, edgeidt>, public messaget {
public:
  typedef graph_weightedt<nodeidt, PropT, edgeidt> grapht;
  typedef typename graph_weightedt<nodeidt, PropT, edgeidt>::edget edget;
  typedef typename graph_weightedt<nodeidt, PropT, edgeidt>::patht patht;
  typedef typename graph_weightedt<nodeidt, PropT, edgeidt>::pathst pathst;
  
  hampath_solver_clingot(message_handlert &_message_handler) : messaget(_message_handler) {
  }

  virtual void solve(typename graph_weightedt<nodeidt, PropT, edgeidt>::patht& path, graph_weightedt<nodeidt, PropT, edgeidt>* gg, nodeidt init, nodeidt final, edgeweightt bound, typename graph_weightedt<nodeidt, PropT, edgeidt>::pathst path_constraints) {

    std::ofstream outfile(OUTFILE);
    if (!outfile.is_open()) throw ("ERROR: cannot open file " OUTFILE);

    //map ids to "dense" set
    typename graph_weightedt<nodeidt, PropT, edgeidt>::nodeidmapt nodeidmap;
    typename graph_weightedt<nodeidt, PropT, edgeidt>::nodest nns;
    gg->get_nodes(nns);
    nodeidt nno = 1;
    for(typename graph_weightedt<nodeidt, PropT, edgeidt>::nodest::iterator n = nns.begin();
        n!=nns.end(); n++, nno++) {
      nodeidmap.insert(std::pair<nodeidt,nodeidt>(*n,nno));
    }
    graph_weighted_intidst<PropT> graph; gg->copy_map_nodeids(graph,nodeidmap);
    init = nodeidmap.at(init);
    final = nodeidmap.at(final);
    typename graph_weightedt<nodeidt, PropT, edgeidt>::pathst newpc;
    for(typename pathst::iterator p = path_constraints.begin(); p!=path_constraints.end(); p++) {
      patht newp = *p;
      grapht::path_map_nodeids(newp,nodeidmap);
      newpc.insert(newp);
    }
    path_constraints = newpc;

    //generate clingo problem
    typename graph_weightedt<nodeidt, PropT, edgeidt>::nodest nodes;
    graph.get_nodes(nodes);
    typename graph_weightedt<nodeidt, PropT, edgeidt>::edgest edges;
    graph.get_edges(edges);

    outfile << "% nodes" << std::endl;
    outfile << "node(1.." << graph.get_node_count() << ")." << std::endl << std::endl;
    outfile << "% edges and costs" << std::endl;
    for(typename graph_weightedt<nodeidt, PropT, edgeidt>::edgest::iterator it=edges.begin(); 
	it!=edges.end(); ++it) {
      outfile << "edge(" << it->first << "," << 
                            it->second << "). ";
      outfile << "cost(" << it->first << "," << 
                            it->second  << "," << (graph.get_edge_label(*it)) << 
         ")." << std::endl;
    }
    outfile << "edge(" << final << "," << init << "). ";
    outfile << "cost(" << final << "," << init << ",0)." << std::endl << std::endl;

    outfile << "% generate" << std::endl;
    outfile << "{ cycle(X,Y) : edge(X,Y) } 1 :- node(X)." << std::endl;
    outfile << "{ cycle(X,Y) : edge(X,Y) } 1 :- node(Y)." << std::endl << std::endl;
    outfile << "reached(Y) :- cycle(1,Y)." << std::endl;
    outfile << "reached(Y) :- cycle(X,Y), reached(X)." << std::endl;
    if(path_constraints.size()>0) {
      outfile << "twotrans(X,Y,Z) :- cycle(X,Y), cycle(Y,Z)." << 
        std::endl << std::endl;
    }
    outfile << ":- node(Y), not reached(Y)." << std::endl << std::endl;

    outfile << "% path constraints" << std::endl;
    for(typename pathst::iterator it=path_constraints.begin();
        it!=path_constraints.end();it++) {
      assert(it->size()==3);
      typename patht::const_iterator p = it->begin();
      outfile << "-twotrans(" << *p; p++;
      outfile << "," << *p; p++;
      outfile << "," << *p << ")." << std::endl;
    }

    outfile  << std::endl << "% Display" << std::endl;
    outfile << "#hide." << std::endl;
    outfile << "#show cycle/2." << std::endl << std::endl;

    outfile << "% Optimize" << std::endl;
    outfile << "#minimize [ cycle(X,Y) : cost(X,Y,C) = C ]." << std::endl;

    outfile.close();

    //run clingo
    assert(system(CLINGO_CMD " " OUTFILE " > " INFILE)>=0); 

    cyclet cycle;
    std::ifstream infile(INFILE);
    if (!infile.is_open()) throw ("ERROR: cannot open file " INFILE);
    if (clingo_parse(infile,&cycle)) throw "ERROR: parse error in clingo result file";
    infile.close();

    //clean up temporary files
    if(!DEBUG_LEVEL) assert(!system(RM_CMD " ./" OUTFILE " ./" INFILE));

    if(cycle.empty()) return;

    //convert cycle list to path
    nodeidt n = init;
    path.push_back(n);
    while(n!=final) {
      cyclet::iterator it = cycle.find(n);
      assert(it!=cycle.end());
      n = it->second;
      path.push_back(n);
    }
    //map back ids
    grapht::reverse_nodeidmap(nodeidmap);
    grapht::path_map_nodeids(path,nodeidmap); 
  }

};

#endif
