/*******************************************************************\

Module: LKH interface

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_HAMPATH_SOLVER_LKH_H
#define CHAINCOVER_HAMPATH_SOLVER_LKH_H

#include "graph_weighted.h"
#include "lkh_parser.h"
#include <util/message.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>

#define LKH_CMD "LKH"
#define OUTPAR "__out__.par"
#define OUTTSP "__out__.tsp"
#define INRES "__in__.res"
#define RM_CMD "rm -f"

#define DEBUG_LEVEL (get_message_handler().get_verbosity()>=9)

template<class PropT>
class hampath_solver_lkht : public hampath_solver_templatet<nodeidt, PropT, edgeidt>, public messaget {
public:
  typedef graph_weightedt<nodeidt, PropT, edgeidt> grapht;
  
  hampath_solver_lkht(message_handlert &_message_handler) : messaget(_message_handler) {
  }

  virtual void solve(typename graph_weightedt<nodeidt, PropT, edgeidt>::patht& path, graph_weightedt<nodeidt, PropT, edgeidt>* gg, nodeidt init, nodeidt final, edgeweightt bound, typename graph_weightedt<nodeidt, PropT, edgeidt>::pathst path_constraints)
{
  if(path_constraints.size()>0) 
    throw "ERROR: LKH does not support path constraints. Use clingo.";

  //map ids to "dense" set
  typename graph_weightedt<nodeidt, PropT, edgeidt>::nodeidmapt nodeidmap;
  typename graph_weightedt<nodeidt, PropT, edgeidt>::nodest nodes;
  gg->get_nodes(nodes);
  nodeidt nno = 1;
  for(typename graph_weightedt<nodeidt, PropT, edgeidt>::nodest::iterator n = nodes.begin();
      n!=nodes.end(); n++, nno++) {
    nodeidmap.insert(std::pair<nodeidt,nodeidt>(*n,nno));
  }
  graph_weighted_intidst<PropT> graph; gg->copy_map_nodeids(graph,nodeidmap);

  //graph.output(std::cout);
  
  init = nodeidmap.at(init);
  final = nodeidmap.at(final);

  //define constants based on given graph
  const int MATRIX_DIMENSION = graph.get_node_count();
  const unsigned WEIGHT_UNCONNECTED = 
    (graph.get_max_weight()+1)*MATRIX_DIMENSION*10;

  //write par file
  std::ofstream outpar(OUTPAR);
  if (!outpar.is_open()) throw ("ERROR: cannot open file " OUTPAR);
  outpar << "PROBLEM_FILE = " << OUTTSP << std::endl;
  outpar << "OPTIMUM = " << bound << std::endl;
  outpar << "MOVE_TYPE = 5" << std::endl;
  outpar << "PATCHING_C = 3" << std::endl;
  outpar << "PATCHING_A = 2" << std::endl;
  outpar << "RUNS = 10" << std::endl;
  outpar << "TOUR_FILE = " << INRES << std::endl;
  outpar << "PRECISION = 1" << std::endl;
  outpar.close();

  //write tsp file
  std::ofstream outtsp(OUTTSP);
  outtsp << "NAME : chaincover" << std::endl;
  outtsp << "TYPE : ATSP" << std::endl;
  outtsp << "DIMENSION : " << MATRIX_DIMENSION << std::endl;
  outtsp << "EDGE_WEIGHT_TYPE : EXPLICIT" << std::endl;
  outtsp << "EDGE_WEIGHT_FORMAT : FULL_MATRIX" << std::endl;
  outtsp << "EDGE_WEIGHT_SECTION :" << std::endl;
  if (!outtsp.is_open()) throw ("ERROR: cannot open file " OUTTSP);
  for (int i = 1; i <= MATRIX_DIMENSION; i++) {
    for (int j = 1; j <= MATRIX_DIMENSION; j++) {
      //get edge weights
      unsigned w = 0;
      typename hampath_solver_lkht<PropT>::grapht::edget edge = 
        typename hampath_solver_lkht<PropT>::grapht::edget(i,j);
      if((i==j) || ((i==final) && (j==init))) w=0;
      else if(!graph.exists_edge(edge)) w = WEIGHT_UNCONNECTED;
      else w = graph.get_edge_label(edge);
      outtsp << w << " ";
    }
    outtsp << std::endl;
  }
  outtsp << "EOF" << std::endl;

  outtsp.close();

  //run LKH
  if(DEBUG_LEVEL) assert(!system(LKH_CMD " " OUTPAR));
  else assert(!system(LKH_CMD " " OUTPAR " > /dev/null 2> /dev/null"));

  //parse LKH output
  lkh_patht lkh_path;
  unsigned cost;
  std::ifstream infile(INRES);
  if (!infile.is_open()) throw ("ERROR: cannot open file " INRES);
  if (lkh_parse(infile,&lkh_path,&cost)) throw "ERROR: parse error in LKH result file";
  infile.close();

  //clean up temporary files
  if(!DEBUG_LEVEL) assert(!system(RM_CMD " ./" OUTPAR " ./" OUTTSP " ./" INRES));

  //extract path from init to final
  if(cost>WEIGHT_UNCONNECTED) return; //no path found
  std::ostringstream out; out << "best tour: ";
  for(lkh_patht::iterator it = lkh_path.begin(); it!=lkh_path.end(); it++) out << *it << " ";
  out << std::endl;
  debug(out.str());
  lkh_patht::iterator it1 = lkh_path.begin();
  while(it1!=lkh_path.end() && *it1!=init) it1++;
  if(it1==lkh_path.end()) throw "ERROR: no init node in TSP cycle";
  lkh_patht::iterator it2 = it1;
  while(it2!=lkh_path.end()) { path.push_back(*it2); it2++; }
  if(it1==lkh_path.begin() && *(--it2)!=final) 
    throw "ERROR: final and init nodes not adjacent in TSP cycle";
  it2 = lkh_path.begin();
  while(it2!=it1) { path.push_back(*it2); it2++; }
  if(it1!=lkh_path.begin() && *(--it2)!=final) 
    throw "ERROR: final and init nodes not adjacent in TSP cycle";

  //map back ids
  grapht::reverse_nodeidmap(nodeidmap);
  grapht::path_map_nodeids(path,nodeidmap);  
}

};

#endif
