/*******************************************************************\

Module: BMC-based reachability engine

Author: Peter Schrammel

\*******************************************************************/

#include <iostream>
#include <util/rename.h>
#include <util/replace_expr.h>
#include <util/irep.h>
#include <langapi/language_util.h>
#include <goto-programs/goto_trace.h>
#include <goto-symex/build_goto_trace.h>
#include <solvers/sat/satcheck_minisat2.h>
#include <solvers/prop/cover_goals.h>
#include <goto-symex/symex_target_equation.h>

#include "reach_bmc.h"
#include "../cbmc_solvers.h"
#include "../bmc.h"
#include "../bv_cbmc.h"


#define MAIN_ID "c::__main__"
#define INPUT_ID "c::__main__::ii"
#define INPUT_ID_BASE "__main__::ii"
#define INPUT_ID_PRETTY "ii"
#define STATE_ID "c::__main__::ss"
#define STATE_ID_BASE "__main__::ss"
#define STATE_ID_PRETTY "ss"
#define INPUT_ID_PTR "c::__main__::i"
#define INPUT_ID_PTR_BASE "__main__::i"
#define INPUT_ID_PTR_PRETTY "i"
#define STATE_ID_PTR "c::__main__::s"
#define STATE_ID_PTR_BASE "__main__::s"
#define STATE_ID_PTR_PRETTY "s"

/******************************************************************************/
/** k-reachability graph construction */
/******************************************************************************/
reach_bmct::prop_edgest reach_bmct::get_k_reach_edges(prop_edgest edges, edgeweightt k) {
  //  if(k<MULTITHREADING_K) 
  //  return get_k_reach_edges_sequential(edges,k);
    return get_k_reach_edges_cover(edges,k);
  //   else return get_k_reach_edges_parallel(edges,k);
}

/** the incremental version */
reach_bmct::prop_edgest reach_bmct::get_k_reach_edges_cover(prop_edgest edges, edgeweightt k) {
  prop_edgest kedges;  
  std::set<propertyt> props1;  
  for(prop_edgest::iterator e=edges.begin();e!=edges.end();e++) {
    props1.insert(e->first);
  }
  for(std::set<propertyt>::iterator p=props1.begin();p!=props1.end();p++) {
    propertiest props2;  
    for(prop_edgest::iterator e=edges.begin();e!=edges.end();e++) {
      if(*p==e->first) {
        props2.push_back(e->second);
      }
    }
    prop_edgest kedgesp = check_k_reach(*p,props2,k);
    kedges.insert(kedgesp.begin(),kedgesp.end());
  }
  return kedges;
}

/** the purely sequential version (not used) */
reach_bmct::prop_edgest reach_bmct::get_k_reach_edges_sequential(prop_edgest edges, edgeweightt k) {
  prop_edgest kedges;  
  for(prop_edgest::iterator e=edges.begin();e!=edges.end();e++) {
    if(check_k_reach(e->first,e->second,k)==REACH) {
      kedges.insert(*e);
    }
  }
  return kedges;
}

/********** some old stuff used to try multi-threaded version ********

#include <thread>

reach_bmct::prop_edgest kedges;
std::mutex kedges_mutex;

void do_check_k_reach(bmct& bmc, reach_bmct reach, reach_bmct::prop_edget e, unsigned k) {    
  if(reach.check_k_reach(bmc,e.first,e.second,k)==reach_bmct::REACH) {
    kedges_mutex.lock();
    kedges.insert(e);
    kedges_mutex.unlock();
  }
}

reach_bmct::prop_edgest reach_bmct::get_k_reach_edges_parallel(prop_edgest edges, unsigned k) {
  kedges.clear();  
  std::thread t[NUM_THREADS];
  bmct* bmc[NUM_THREADS];
  prop_edgest::iterator e=edges.begin();
  while(true) {
    unsigned n;
      for(n=0;n<NUM_THREADS;n++) {
        bmc[n] = new bmct(*context->options, *context->symbol_table, *context->ui_message_handler);
        t[n] = std::thread(do_check_k_reach,*(bmc[n]),*this,*e,k);
        e++;
        if(e==edges.end()) break;
      } 
      for(unsigned i=0;i<n;i++) {
        t[i].join();
      }
      for(unsigned i=0;i<n;i++) {
        delete bmc[i];
      }
    if(e==edges.end()) break;
  } 
  return kedges; 
}
*/

/******************************************************************************/
/*** helpers for generating the programs ***/
/******************************************************************************/

/** copies the property entry code of a given property function to a goto program */
void reach_bmct::add_prop_entry(goto_programt &goto_program,irep_idt prop_name,bool as_assertion,symbol_exprt ivar, symbol_exprt svar) {
  const goto_functionst::goto_functiont *prop;
  goto_functionst::function_mapt::const_iterator prop_it = context->goto_functions->function_map.find(prop_name);
  if(prop_it==context->goto_functions->function_map.end()) 
    throw "ERROR: function"+id2string(prop_name)+" not found";
  prop = &(prop_it->second);

  //argument names
  irep_idt iname = prop->type.arguments()[0].get_identifier();
  irep_idt sname = prop->type.arguments()[1].get_identifier();
  //check types and adapt to order of arguments
  std::string type_error = "argument types of function "+id2string(prop_name)+" do not match configuration";
  typet t1 = prop->type.arguments()[0].type(); 
  if(t1.id()!=ID_pointer) throw type_error;
  pointer_typet pt1 = to_pointer_type(t1);
  symbol_typet spt1 = to_symbol_type(pt1.subtype());
  dstring type1 = id2string(spt1.get_identifier());
  typet t2 = prop->type.arguments()[1].type();
  pointer_typet pt2 = to_pointer_type(t2);
  if(t2.id()!=ID_pointer) throw type_error;
  symbol_typet spt2 = to_symbol_type(pt2.subtype());
  dstring type2 = id2string(spt2.get_identifier());
  if(type1==context->itype && type2==context->stype) ; //alright 
  else if(type2==context->itype && type1==context->stype) { //swap
    irep_idt tmp = sname;
    sname = iname;
    iname = tmp;
  }
  else throw type_error;

  // Definitions for mapping between the two programs
  typedef std::map<goto_programt::const_targett, goto_programt::targett> 
    targets_mappingt;
  targets_mappingt targets_mapping;

  // find last assume statement where we stop copying
  goto_programt::instructionst::const_iterator break_instr = 
    (prop->body).instructions.end(); 
  if(as_assertion) {
    for(goto_programt::instructionst::const_iterator 
	  it=(prop->body).instructions.begin(); 
	it!=(prop->body).instructions.end(); it++) {
      if(it->is_assume()) {
        break_instr = it; break_instr++;
        break; //break at first assumption (=property trigger)
      }
    }
  }

#if 0
  if(break_instr != (prop->body).instructions.end())
    std::cout << "break_instr: " << break_instr->location_number << std::endl;
  else std::cout << "break_instr: end" << std::endl;
#endif

  // collect targets and copy
  for(goto_programt::instructionst::const_iterator 
      it=(prop->body).instructions.begin(); 
       it!=break_instr; it++) {
    goto_programt::targett new_instruction=goto_program.add_instruction();
    targets_mapping[it]=new_instruction;
    *new_instruction=*it;
    rename(new_instruction->guard,sname,svar.get_identifier());
    rename(new_instruction->guard,iname,ivar.get_identifier());
    rename(new_instruction->code,sname,svar.get_identifier());
    rename(new_instruction->code,iname,ivar.get_identifier());
    if(it->is_assert() || it->is_end_function()) {
      new_instruction->make_skip(); //remove assertions
    }
    if(it->is_assume()) {
      if(as_assertion) {
	new_instruction->type = ASSERT;
        new_instruction->guard = not_exprt(new_instruction->guard);
      }
    }
  }

  // update targets
  for(goto_programt::instructionst::iterator
      it=goto_program.instructions.begin();
      it!=goto_program.instructions.end();
      it++)
  {
    for(goto_programt::instructiont::targetst::iterator
        t_it=it->targets.begin();
        t_it!=it->targets.end();
        t_it++)
    {
      targets_mappingt::iterator
        m_target_it=targets_mapping.find(*t_it);

      if(m_target_it!=targets_mapping.end())
        *t_it=m_target_it->second;
    }
  }

  goto_program.compute_incoming_edges();
  goto_program.compute_target_numbers();

#if 0
  goto_program.output(namespacet(*context->symbol_table),"",std::cout);
#endif
}

/** adds a function call with one or two arguments to a goto program 
     (the order of arguments is adapted to the formal arguments) */
void reach_bmct::add_function_call(goto_programt &goto_program,const irep_idt name,exprt arg1, irep_idt argtype1, exprt arg2, irep_idt argtype2) {
  //check function
  goto_functionst::function_mapt::const_iterator fun_it = context->goto_functions->function_map.find(name);
  if(fun_it==context->goto_functions->function_map.end()) 
    throw "ERROR: function "+id2string(name)+" not found";
  if(fun_it->second.type.arguments().size()==0 || 
      fun_it->second.type.arguments().size()>=3) 
    throw "expected one or two arguments for function "+id2string(name);

  //check argument type
  std::string type_error = "argument types of function "+id2string(name)+" do not match configuration";
  typet t1 = fun_it->second.type.arguments()[0].type(); 
  if(t1.id()!=ID_pointer) throw type_error;
  pointer_typet pt1 = to_pointer_type(t1);
  symbol_typet spt1 = to_symbol_type(pt1.subtype());
  dstring type1 = id2string(spt1.get_identifier());
  if(fun_it->second.type.arguments().size()==2) {
    typet t2 = fun_it->second.type.arguments()[1].type(); 
    if(t2.id()!=ID_pointer) throw type_error;
    pointer_typet pt2 = to_pointer_type(t2);
    symbol_typet spt2 = to_symbol_type(pt2.subtype());
    dstring type2 = id2string(spt2.get_identifier());
    if(type1==argtype1 && type2==argtype2) ; // alright
    else if(type1==argtype2 && type2==argtype1) { //swap
      exprt tmp = arg1;
      arg1 = arg2;
      arg2 = tmp;
    }
    else throw type_error;
  }
  else { //one argument
    if(type1==argtype1) ; // alright
    else if(type1==argtype2) { //take second argument
      arg1 = arg2;
    }
    else throw type_error;
  }

  symbol_exprt expr;
  expr.set_identifier(name);
  goto_programt::targett i_call = goto_program.add_instruction(FUNCTION_CALL);
  code_function_callt c_call = code_function_callt();
  c_call.function() = expr;
  c_call.arguments().push_back(arg1);
  if(fun_it->second.type.arguments().size()==2) 
    c_call.arguments().push_back(arg2);
  i_call->code.swap(c_call);
}

/** adds a variable declaration to a goto program */
void reach_bmct::add_var_declaration(goto_programt &goto_program,symbol_exprt& var, irep_idt name, irep_idt pretty_name, irep_idt base_name, typet type) {
  var.set_identifier(name);
  var.type()=type;
  symbolt sym;
  sym.clear();
  sym.name = name;
  sym.base_name = base_name;
  sym.pretty_name = pretty_name;
  sym.type = type;
  context->symbol_table->add(sym); //does nothing if symbol exists
  goto_programt::targett decl = goto_program.add_instruction(DECL);
  decl->code = code_declt(var);  
}

/** adds an assignment of the inputs with non-deterministic value to a goto program */
void reach_bmct::add_nondet_assign_inputs(goto_programt &goto_program,exprt iexpr, exprt iiexpr) {
  goto_programt::targett i_assign = goto_program.add_instruction(ASSIGN);
  i_assign->code = code_assignt(dereference_exprt(iexpr), side_effect_expr_nondett(iiexpr.type()));
}

/** adds an assignment of the state to the given values to a goto program */
void reach_bmct::add_assign_state(goto_programt &goto_program, symbol_exprt ssexpr, valuationt v) {
  for(valuationt::iterator it = v.begin(); it!=v.end(); it++) {
    goto_programt::targett i_assign = goto_program.add_instruction(ASSIGN);
    i_assign->code = code_assignt(ssexpr,it->second);
  }
}

/******************************************************************************/
/** checks k-reachability from prop1 to a list of props */
/******************************************************************************/
reach_bmct::prop_edgest reach_bmct::check_k_reach(propertyt prop1, propertiest props,
edgeweightt k)
{
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false;
  bmc.generate_error_trace = false;
    
  /*** construct program ***/

  //  if(k==0 && prop1!=context->init) return prop_edgest(); 

  std::ostringstream out;
  for(propertiest::iterator it = props.begin(); it != props.end(); it++) {
    out << "--- check " << k << "-reach for (" << prop1 << "," << *it << ") ---" << std::endl;
  }
  bmc.debug(out.str());

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisations
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  //assumption calls
  add_function_call(main,context->state_assume,
		    sexpr,context->stype,sexpr,context->stype); //2nd=dummy arg

 //prop from
  if(prop1!=context->init) add_prop_entry(main,prop1,false,iexpr,sexpr);
  else add_function_call(main,context->init,
         iexpr,context->itype,sexpr,context->stype);

  //unroll
  for(edgeweightt i=0;i<k;i++) {
    //nondet assignment of inputs 
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
    //step call
    add_function_call(main,context->step,
      iexpr,context->itype,sexpr,context->stype);
  }
  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
 
  //props to
  for(propertiest::iterator it = props.begin(); it != props.end(); it++) {
    add_prop_entry(main,*it,true,iexpr,sexpr);//as assertion
  }

  main.add_instruction(END_FUNCTION);

#if 0
  //  main.compute_location_numbers();
  main.output(namespacet(*context->symbol_table),"",std::cout);
#endif

  //run symex
  bmc.run_symex(*context->goto_functions,main);  

  // convert
  bmc.equation.convert(prop_conv);

  // collect _all_ goals in `goal_map'
  typedef std::map<goto_programt::const_targett, exprt::operandst> goal_mapt;
  goal_mapt goal_map;
  typedef std::map<goto_programt::const_targett, propertyt> prop_mapt;
  prop_mapt prop_map;
  
  propertiest::iterator pit = props.begin();
  forall_goto_program_instructions(i_it, main) {
    if(i_it->is_assert()) {
      //      goal_map[i_it]=bvt();
      prop_map[i_it]=*pit;
      ++pit;
    }
  }

  // get the conditions for these goals from formula
  for(symex_target_equationt::SSA_stepst::iterator
      it=bmc.equation.SSA_steps.begin();
      it!=bmc.equation.SSA_steps.end();
      it++)
  {
    if(it->is_assert())
    {
      goal_map[it->source.pc].push_back(literal_exprt(!it->cond_literal));
    }
  }

  
  // try to cover those
  cover_goalst cover_goals(prop_conv);
  cover_goals.set_message_handler(*context->ui_message_handler);

  for(goal_mapt::const_iterator
      it=goal_map.begin();
      it!=goal_map.end();
      it++)
  {
    // the following is TRUE if the bv is empty
    literalt p=prop_conv.convert(disjunction(it->second));
    cover_goals.add(p);
  }

  out.clear();
  out << "Total number of coverage goals: " << cover_goals.size() << std::endl;
  bmc.debug(out.str());

  cover_goals();

  prop_edgest result;
  std::list<cover_goalst::cover_goalt>::const_iterator g_it=
    cover_goals.goals.begin();
      
  out.clear();
  for(prop_mapt::const_iterator
      it=prop_map.begin();
      it!=prop_map.end();
      it++, g_it++)
  {
      if(g_it->covered) {
        out << it->second << ": covered" << std::endl;
        result.insert(prop_edget(prop1,it->second));
      }
      else { out << it->second << ": not covered" << std::endl; }
  }
  bmc.debug(out.str());

  return result;
}

/******************************************************************************/
/** checks k-reachability from prop1 to prop2 (not used) */
/******************************************************************************/
reach_bmct::resultt reach_bmct::check_k_reach(propertyt prop1, propertyt prop2,
edgeweightt k)
{
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false;
  bmc.generate_error_trace = false;

  if(k==0 && prop1!=context->init) return UNREACH; 

  std::ostringstream out;
  out << "--- check " << k << "-reach for (" << prop1 << "," << prop2 << ") ---" << std::endl;
  bmc.debug(out.str());

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisations
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  //assumption calls
  add_function_call(main,context->state_assume,
    sexpr,context->stype,sexpr,context->stype); //2nd = dummy arg

 //prop from
  if(prop1!=context->init) add_prop_entry(main,prop1,false,iexpr,sexpr);
  else add_function_call(main,context->init,
         iexpr,context->itype,sexpr,context->stype);

  //unroll
  for(edgeweightt i=0;i<k;i++) {
    //nondet assignment of inputs 
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
    //step call
    add_function_call(main,context->step,
      iexpr,context->itype,sexpr,context->stype);
  }
  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
 
  //prop to
  add_prop_entry(main,prop2,true,iexpr,sexpr);//as assertion

  main.add_instruction(END_FUNCTION);

  // do actual BMC
  bool res = bmc.run(*context->goto_functions,main);

  if(res) return REACH; else return UNREACH; 
}

/******************************************************************************/
/** checks whether a given path is feasible */
/******************************************************************************/
bool reach_bmct::check_path(reach_bmct::prop_patht path, reach_bmct::pathweightst pathweights) {
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false;
  bmc.generate_error_trace = false;

  bmc.debug("--- checking path ---");

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisation
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  reach_bmct::pathweightst::iterator w=pathweights.begin();
  reach_bmct::prop_patht::iterator n=path.begin();

  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
  //start
  if(*n==context->init) {
    add_function_call(main,context->init,
      iexpr,context->itype,sexpr,context->stype);
  }
  else {
    add_function_call(main,context->state_assume,
      sexpr,context->stype,sexpr,context->stype); // 2nd = dummy arg
    add_prop_entry(main,*n,false,iexpr,sexpr);
  }
  n++;
  while(w!=pathweights.end()) {
    //unroll
    for(edgeweightt i=0;i<*w;i++) {
      //nondet assignment of inputs 
      add_nondet_assign_inputs(main,iexpr,iiexpr);
      add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
      //step call
      add_function_call(main,context->step,
        iexpr,context->itype,sexpr,context->stype);
    }
    //nondet assignment of inputs 
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		      iexpr,context->itype,sexpr,context->stype);
    //prop entry 
    if(++w==pathweights.end()) add_prop_entry(main,*n,true,iexpr,sexpr);
    else add_prop_entry(main,*n,false,iexpr,sexpr);
    n++;
  }
  main.add_instruction(END_FUNCTION);

  // do actual BMC
  bool res = bmc.run(*context->goto_functions,main);

  return res;
  }

/******************************************************************************/
/** checks whether there is a path from (a given state) in a property to another property in a given number of steps and returns the end state of the path, if so */
/******************************************************************************/
bool reach_bmct::check_steps(propertyt beginprop, valuationt beginstate, propertyt endprop, edgeweightt steps, valuationt& endstate, testvectort &chain) {
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false;
  bmc.generate_error_trace = true;

  bmc.debug("--- checking path ---");

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  add_assign_state(main,ssexpr,beginstate);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisation
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
  //start
  if(beginprop==context->init) {
    add_function_call(main,context->init,
      iexpr,context->itype,sexpr,context->stype);
  }
  else {
    add_function_call(main,context->state_assume,
      sexpr,context->stype,sexpr,context->stype); //2nd = dummy arg
    add_prop_entry(main,beginprop,false,iexpr,sexpr);
  }

  //unroll
  for(edgeweightt i=0;i<steps;i++) {
    //nondet assignment of inputs
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
    //step call
    add_function_call(main,context->step,
      iexpr,context->itype,sexpr,context->stype);
  }
  //nondet assignment of inputs
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);
  //end
  add_prop_entry(main,endprop,true,iexpr,sexpr);
  main.add_instruction(END_FUNCTION);
  namespacet ns(*context->symbol_table);

  /*  std::ostringstream out;
  main.output(ns,MAIN_ID,out);
  bmc.debug(out.str());*/

  // do actual BMC
  bool res = bmc.run(*context->goto_functions,main);

  if(res) { //build chain
    endstate.clear();
    for(goto_tracet::stepst::iterator it = bmc.goto_trace.steps.end();
        it!=bmc.goto_trace.steps.begin();) {
      it--;
      if(it->pc->is_assign())
      {
        irep_idt identifier=it->lhs_object.get_identifier();
        if(identifier==ssexpr.get_identifier()) {
          endstate.insert(valuationt::value_type(it->lhs_object, it->lhs_object_value));
	}

        if(identifier==INPUT_ID && it->full_lhs_value.id()==ID_struct) {
          valuationt v;
          v.insert(valuationt::value_type(it->lhs_object, it->full_lhs_value));
          chain.steps.push_front(v);
        }
      }
    }
    chain.steps.pop_back();
  }

  return res;
}

/******************************************************************************/
/** computes a chain by property-to-property reachability calls chaining on concrete states */
/******************************************************************************/
void reach_bmct::repair_chain(prop_patht proppath, pathweightst& pathweights, prop_patht& failed_path, edgeweightt reach_bound, testvectort &chain) {
  assert(failed_path.size()>=3);
  prop_patht partproppath;
  pathweightst partpathweights;
  pathweightst::iterator w = pathweights.begin();
  prop_patht::iterator p = proppath.begin();
  partproppath.push_back(*p);
  for(; p != proppath.end() && *p != failed_path.front(); p++, w++) {
    partproppath.push_back(*p);
    partpathweights.push_back(*w);
  }
  valuationt state;
  p = failed_path.begin();
  while(true) {
    prop_patht::iterator lastp = p;
    p++;
    partpathweights.push_back(*w);
    partproppath.push_back(*p);
    if(p==failed_path.end()) break;
    bool res = false;
    testvectort partchain;
    while(!res) {
      std::ostringstream out;
      out << "checking path: " << (*lastp) << " --" << (*w) << "--> " << (*p);
      debug(out.str());

      partchain.steps.clear();
      res = check_steps(*lastp, state, *p, *w, state, partchain);
      if(!res) {
	(*w)++;
        if(*w>reach_bound) {
          chain.steps.clear();
 
          //check path from init for this section 
          //          (concrete chaining alone is not complete)
          pathweightst::iterator ww = partpathweights.end(); ww--; //old weight
          while(!res) {
            res = check_path_chain(partproppath,partpathweights,state,partchain);
            if(!res) { 
              (*ww)++;
              if(*ww>reach_bound) {
                      //now we know that this section is infeasible up to bound
                //3-subpath that causes the problem
                assert(lastp!=proppath.begin()); 
                propertyt p3 = *p;
                propertyt p2 = *(--p);
                propertyt p1 = *(--p);
                failed_path.clear();  
  	        failed_path.push_back(p1);
	        failed_path.push_back(p2);
	        failed_path.push_back(p3);
                return;
	      }
	    }
	  }
          *w = *ww; //replace by updated weight (check from init successful)
	}
      }
      else { //replace by updated weight (concrete chaining successful)
        partpathweights.pop_back();
        partpathweights.push_back(*w);
      }
    }
    chain.steps.insert(chain.steps.end(),partchain.steps.begin(),partchain.steps.end());
    w++;
  }
}

/******************************************************************************/
/** computes a chain along the given path and returns last state */
/******************************************************************************/
bool reach_bmct::check_path_chain(reach_bmct::prop_patht path, reach_bmct::pathweightst pathweights, valuationt& endstate, testvectort& chain) {
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false; //true;
  bmc.generate_error_trace = true;

  bmc.debug("--- computing chain ---");

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisation
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  //init call
  add_function_call(main,context->init,
    iexpr,context->itype,sexpr,context->stype);
  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);

  reach_bmct::pathweightst::iterator w=pathweights.begin();
  reach_bmct::prop_patht::iterator n=path.begin();
  n++; //jump over init
  while(w!=pathweights.end()) {
    //unroll
    for(edgeweightt i=0;i<*w;i++) {
      //nondet assignment of inputs 
      add_nondet_assign_inputs(main,iexpr,iiexpr);
      add_function_call(main,context->input_assume,
			iexpr,context->itype,sexpr,context->stype);
      //step call
      add_function_call(main,context->step,
			iexpr,context->itype,sexpr,context->stype);
    }
    //nondet assignment of inputs 
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		      iexpr,context->itype,sexpr,context->stype);
    //prop entry 
    if(++w==pathweights.end()) add_prop_entry(main,*n,true,iexpr,sexpr);
    else add_prop_entry(main,*n,false,iexpr,sexpr);
    n++;
  }
  main.add_instruction(END_FUNCTION);

  // do actual BMC
  bool res = bmc.run(*context->goto_functions,main);

  if(res) { //build chain
    endstate.clear();
    for(goto_tracet::stepst::iterator it = bmc.goto_trace.steps.end();
        it!=bmc.goto_trace.steps.begin();) {
      it--;
      if(it->pc->is_assign())
      {
        irep_idt identifier=it->lhs_object.get_identifier();
        if(identifier==ssexpr.get_identifier()) {
          endstate.insert(valuationt::value_type(it->lhs_object, it->lhs_object_value));
	}

        if(identifier==INPUT_ID && it->full_lhs_value.id()==ID_struct) {
          valuationt v;
          v.insert(valuationt::value_type(it->lhs_object, it->full_lhs_value));
          chain.steps.push_front(v);
        }
      }
    }
    chain.steps.pop_back();
  }

  return res;
}


/******************************************************************************/
/** computes the test case along the given path (not used) */
/******************************************************************************/
bool reach_bmct::get_chain(testvectort& chain, reach_bmct::prop_patht& failed_path, reach_bmct::prop_patht path, reach_bmct::pathweightst pathweights) {
  //get solver
  cbmc_solverst cbmc_solvers(*context->options, *context->symbol_table, 
           *context->ui_message_handler);
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();
  bmct bmc(*context->options, *context->symbol_table, 
           *context->ui_message_handler,prop_conv);
  bmc.set_ui(context->ui);
  //  bmc.core_enabled = false; //true;
  bmc.generate_error_trace = true;

  bmc.debug("--- computing chain ---");

  goto_functionst::goto_functiont fmain;
  goto_programt main;
  fmain.body = main;

  typet itype = context->symbol_table->lookup(context->itype).type;
  typet stype = context->symbol_table->lookup(context->stype).type;

  //variable declarations
  symbol_exprt ssexpr;
  add_var_declaration(main,ssexpr,STATE_ID,STATE_ID_BASE,STATE_ID_PRETTY,stype);
  symbol_exprt sexpr;
  add_var_declaration(main,sexpr,STATE_ID_PTR,STATE_ID_PTR_BASE,STATE_ID_PTR_PRETTY,pointer_typet(stype));
  symbol_exprt iiexpr;
  add_var_declaration(main,iiexpr,INPUT_ID,INPUT_ID_BASE,INPUT_ID_PRETTY,itype);
  symbol_exprt iexpr;
  add_var_declaration(main,iexpr,INPUT_ID_PTR,INPUT_ID_PTR_BASE,INPUT_ID_PTR_PRETTY,pointer_typet(itype));

  //initialisation
  goto_programt::targett i_sassign = main.add_instruction(ASSIGN);
  i_sassign->code = code_assignt(sexpr, address_of_exprt(ssexpr));
  goto_programt::targett i_iassign = main.add_instruction(ASSIGN);
  i_iassign->code = code_assignt(iexpr, address_of_exprt(iiexpr));

  //init call
  add_function_call(main,context->init,
    iexpr,context->itype,sexpr,context->stype);
  //nondet assignment of inputs 
  add_nondet_assign_inputs(main,iexpr,iiexpr);
  add_function_call(main,context->input_assume,
		    iexpr,context->itype,sexpr,context->stype);

  reach_bmct::pathweightst::iterator w=pathweights.begin();
  reach_bmct::prop_patht::iterator n=path.begin();
  n++; //jump over init
  while(w!=pathweights.end()) {
    //unroll
    for(edgeweightt i=0;i<*w;i++) {
      //nondet assignment of inputs 
      add_nondet_assign_inputs(main,iexpr,iiexpr);
      add_function_call(main,context->input_assume,
 		    iexpr,context->itype,sexpr,context->stype);
      //step call
      add_function_call(main,context->step,
        iexpr,context->itype,sexpr,context->stype);
    }
    //nondet assignment of inputs 
    add_nondet_assign_inputs(main,iexpr,iiexpr);
    add_function_call(main,context->input_assume,
		      iexpr,context->itype,sexpr,context->stype);
    //prop entry 
    if(++w==pathweights.end()) add_prop_entry(main,*n,true,iexpr,sexpr);
    else add_prop_entry(main,*n,false,iexpr,sexpr);
    n++;
  }
  main.add_instruction(END_FUNCTION);

  // do actual BMC
  bool res = bmc.run(*context->goto_functions,main);

  if(res) { //build chain
    for(goto_tracet::stepst::iterator it = bmc.goto_trace.steps.begin();
        it!=bmc.goto_trace.steps.end();++it) {
      if(it->pc->is_assign())
      {
        irep_idt identifier=it->lhs_object.get_identifier();
        if(identifier==INPUT_ID && it->full_lhs_value.id()==ID_struct) {
          valuationt v;
          v.insert(valuationt::value_type(it->lhs_object, it->full_lhs_value));
          chain.steps.push_back(v);
        }
      }
    }
    if(!chain.steps.empty()) chain.steps.pop_back();
  }
  else { 

    /******* some old stuff to use unsat cores to restrict the failed path ******
        //build failed_path
        //  extract failed path from unsat core (not very efficient)
        std::set<irep_idt> funcs;
    for(symex_target_equationt::SSA_stepst::iterator it = 
        bmc.equation.SSA_steps.begin();
        it != bmc.equation.SSA_steps.end();
        ++it) { 
      //  if(it->is_assume()) {
          exprt asserted_expr = it->asserted_expr;
          if(!bmc.solver->in_core(asserted_expr)) { 
	  //	std::cout << "asserted_expr: " << asserted_expr << std::endl;
        // relate to property
	  goto_programt::const_targett instr =  it->source.pc;
	  irep_idt func = instr->function;
  	  funcs.insert(func);
	//      failed_path.push_back(func);
	  //	}
      }
      } 
    std::cout <<  "failed_path = { ";
    for(prop_patht::iterator it = path.begin(); 
	it != path.end(); ++it) {
      if(funcs.find(*it) != funcs.end()) {
        std::cout << *it << " ";
        failed_path.push_back(*it);
      }
    }
    std::cout << "}" << std::endl; 
*/

    failed_path = path; //take whole path
  }
  return res;
}

/******************************************************************************/
/** creates a test harness for the given chain (not fully implemented yet) */
/******************************************************************************/
void reach_bmct::output_harness(testvectort chain, reach_bmct::prop_patht path, reach_bmct::pathweightst pathweights, std::string outfile) {

  //TODO
  throw "NOT IMPLEMENTED.";

  std::ofstream out;
  out.open(outfile.c_str());
  if(!out.is_open()) throw "ABORTED. file '"+outfile+"' could not be opened";

  out << "#define NO_CBMC" << std::endl << std::endl;
  out << "#include \"" << context->harness_incfile << "\"" << std::endl << std::endl;
  out << "void main() {" << std::endl
      << "  unsigned covered = 0; " << std::endl
      << "  unsigned passed = 0; " << std::endl
      << "  unsigned failed = 0; " << std::endl;
  std::string str_itype = id2string(context->itype).substr(7,id2string(context->itype).length()-7);
  std::string str_stype = id2string(context->stype).substr(7,id2string(context->stype).length()-7);
  //declarations
  out << "  struct " << str_stype << " ss;" << std::endl
      << "  struct " << str_itype << " ii;" << std::endl
      << "  struct " << str_stype << "* s = &ss;" << std::endl
      << "  struct " << str_itype << "* i = &ii;" << std::endl;
  //init call
  out << "  " << id2string(context->init).substr(3,id2string(context->init).length()-3) << 
                 "(s);" << std::endl; 
  //steps
  reach_bmct::pathweightst::iterator w=pathweights.begin();
  reach_bmct::prop_patht::iterator n=path.begin();
  n++; //jump over init
  edgeweightt w_current = *w;
  for(stepst::iterator c=chain.steps.begin(); 
      c!=chain.steps.end();c++) {
    //assign inputs
    assert(c->size()==1);
    valuationt::iterator v = c->begin();
    assert(to_symbol_expr(v->first).get_identifier()=="c::__main__::ii");
    struct_exprt vv = to_struct_expr(v->second);
    struct_typet vvt = to_struct_type(v->second.type());
    for(unsigned i=0;i<vv.operands().size();i++) {
      out << "  ii." << id2string(vvt.components()[i].get(ID_name)) << " = " 
          << from_expr(vv.operands()[i]) << ";" << std::endl;
    }
    if(w_current==0) {
      //TODO: insert property code
      //check coverage
      std::string assumption;      
      std::string assertion;
/*      if(*n=="c::prop1") {
        assumption = "(s->mode==1) && (s->speed==1) && (s->button==1) && (i->signal==I_DEC)";
        assertion = "s->speed==1";
      }
*/
      if(*n!="c::final") {
      out << "  if(" << assumption << ") {" << std::endl
          << "    covered++;" << std::endl;

      //step call
      out << "    " << id2string(context->step).substr(3,id2string(context->step).length()-3) << "(i,s);" << std::endl;

      //check verdict
      out << "    if(" << assertion << ") {" << std::endl
          << "      passed++;" << std::endl
          << "      printf(\"" << *n << " passed\\n\");" << std::endl
          << "    }" << std::endl; 
      out << "    else {" << std::endl
          << "      failed++;" << std::endl
          << "      printf(\"" << *n << " failed\\n\");" << std::endl
          << "    }" << std::endl; 
      out << "  }" << std::endl
          << "  else {" << std::endl
          << "    printf(\"" << *n << " not covered\\n\");" << std::endl
          << "    " << id2string(context->step).substr(3,id2string(context->step).length()-3) << "(i,s);" << std::endl
          << "  }" << std::endl;
      
      w_current = *(++w);
      n++;
      }
    }
    else {
      //step call
      out << "  " << id2string(context->step).substr(3,id2string(context->step).length()-3) << "(i,s);" << std::endl;
    }
    w_current--;
  }
  //TODO: check final reached
/*  out << "  if((s->mode==0) && (s->speed==0) && (s->button==0))" << std::endl
      << "    printf(\"final reached\\n\");" << std::endl
      << "  else printf(\"final not reached\\n\");" << std::endl;
*/
  out << "  printf(\"\\nStatistics:\\n\");" << std::endl
      << "  printf(\"  total properties: " << 
                           (path.size()-2) << "\\n\");" << std::endl
      << "  printf(\"  covered:          %d\\n\",covered);" << std::endl 
      << "  printf(\"  passed:           %d\\n\",passed);" << std::endl
      << "  printf(\"  failed:           %d\\n\",failed);" << std::endl;
  out << "}" << std::endl;
  out.close();
}

void reach_bmct::output_chain_file(testvectort chain, std::string outfile) {
  std::ofstream out;
  out.open(outfile.c_str());
  if(!out.is_open()) throw "ABORTED. file '"+outfile+"' could not be opened";

  out << "#define CHAIN_LENGTH " << chain.steps.size() << std::endl << std::endl;
  std::string str_itype = id2string(context->itype).substr(7,id2string(context->itype).length()-7);
  out << "struct " << str_itype << " io[CHAIN_LENGTH] = {" << std::endl;
  output_chain(chain,out);
  out << "};" << std::endl;
  out.close();
}
