#include <climits>

#include <util/time_stopping.h>
#include <goto-symex/slice.h>

#include "../bmc.h"
#include "../cbmc_solvers.h"

#include "testchain_sat.h"

/*******************************************************************\

Function: bmct::do_chain_test_goals

  Inputs:

 Outputs:

 Purpose: compute test chain

\*******************************************************************/

void testchain_satt::compute(goto_functionst &goto_functions, 
                             symbol_tablet &symbol_table) 
{
  absolute_timet time_start(current_time());
  //get solver
  cbmc_solverst cbmc_solvers(options, symbol_table, get_message_handler());
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();

  bmct bmc(options, symbol_table, get_message_handler(), prop_conv);
 
  config->parse();
  propertiest test_goals0 = config->properties; //copy

  //count goals
  bool final_found = false;
  goal_mapt goal_map;
  Forall_goto_functions(f_it, goto_functions) 
  {
    Forall_goto_program_instructions(i_it, f_it->second.body) 
    {
      if(i_it->is_assert()) 
      {
        irep_idt comment = i_it->location.get_comment();
        if(id2string(comment)=="") continue;
        if(comment==config->final) 
	{
	  final_found = true;
	  i_it->guard.make_false();
	  goal_map[i_it->location.get_property_id()]=goalt(*i_it);
	  continue;
	}
        std::set<propertyt>::iterator tg = 
          test_goals0.find(comment);
        if(tg!=test_goals0.end())
	{
	  i_it->guard.make_false();
          goal_map[i_it->location.get_property_id()]=goalt(*i_it);
          test_goals0.erase(tg);
	}
        else i_it->make_skip();
      }
    }
  } 
  if(!final_found) test_goals0.insert(config->final);
  if(!test_goals0.empty()) 
                          //some test goals in the configuration 
  {                       //  were not found in the code
    std::string msg = "test goals not found: ";
    for(std::set<propertyt>::iterator it = test_goals0.begin();
        it != test_goals0.end(); it++) 
    {
      msg += id2string(*it) + " ";
    }
    throw msg;
  }
  status() << "Number of test goals (including final): " 
	   << (goal_map.size()) << eom;

  //find entry point
  goto_functionst::function_mapt::const_iterator it=
      goto_functions.function_map.find(config->main);
  if(it==goto_functions.function_map.end())
      throw "main symbol not found; please set an entry point";
  const goto_programt &goto_program=it->second.body;

  //setup incremental BMC
  bmc.symex.set_message_handler(get_message_handler());
  bmc.symex.options=options;
  bmc.symex.last_location.make_nil();
  bmc.setup_unwind();
  if(!bmc.symex.max_unwind_is_set && config->chain_bound>0) 
  {
    bmc.symex.set_unwind_limit(config->chain_bound);
  }
  if(bmc.symex.incr_max_unwind<UINT_MAX && config->chain_bound>0) 
    bmc.symex.incr_max_unwind = config->chain_bound;

  //incremental BMC loop
  bool chain_found = false;
  bool symex_done = false;
  goto_symext::statet symex_state;
  unsigned unwindings = 0;
  while(!chain_found && !symex_done) 
  { 
    unwindings++;
    symex_done = bmc.symex(symex_state,goto_functions,goto_program);
    undo_slice(bmc.equation); //undo all previous slicings
    statistics() << "size of program expression: "
		 << bmc.equation.SSA_steps.size()
		 << " steps" << eom;
    slice(bmc.equation);
    //convert  
    bmc.do_unwind_module(prop_conv);
    bmc.equation.convert_guards(prop_conv);
    bmc.equation.convert_assignments(prop_conv);
    bmc.equation.convert_decls(prop_conv);
    bmc.equation.convert_assumptions(prop_conv);

    if(!add_chain_constraints(goto_functions,goal_map,bmc)) 
    {
      status() << "Continuing because there are unreachable test goals..." 
               << eom;
      continue;
    }

    //run decision procedure
    absolute_timet sat_start=current_time();
    decision_proceduret::resultt dec_result=prop_conv.dec_solve();
    {
      absolute_timet sat_stop=current_time();
      status() << "Runtime decision procedure: "
	       << (sat_stop-sat_start) << "s" << eom;
    }

    chain_found = (dec_result==decision_proceduret::D_SATISFIABLE);
  }
  if(chain_found) 
  {
    status() << "Chain found (" << unwindings << " unwindings):" << eom;
    add_testvector(goto_functions,bmc);
  }
  else status() << "No chain found after " << unwindings << 
         " unwindings" << eom;

  statistics() << "Total time: " << (current_time()-time_start) << eom;
}


bool testchain_satt::add_chain_constraints(goto_functionst &goto_functions,
					   goal_mapt &goal_map,
					   bmct &bmc) 
{
  //reset conjuncts in goal map
  for(goal_mapt::iterator it=goal_map.begin();  it!=goal_map.end(); it++)
  {
    it->second.conjuncts.clear();
  }

  // get the conditions for these goals from formula
  exprt::operandst assumptions;
  unsigned property_counter=0;
  for(symex_target_equationt::SSA_stepst::iterator
      it=bmc.equation.SSA_steps.begin();
      it!=bmc.equation.SSA_steps.end();
      it++)
  {
    if(it->is_assume())
    {
      assumptions.push_back(literal_exprt(it->cond_literal));
    }
    if(it->is_assert())
    {
      irep_idt property_id;

      if(it->source.pc->is_assert())
        property_id=it->source.pc->location.get_property_id();
      else
      {
        // need new property ID, say for an unwinding assertion
        property_counter++;
        property_id=i2string(property_counter);
        goal_map[property_id].description=it->comment;
      }

#if 0
      std::cout << "goal " << property_id << ": " 
                << from_expr(bmc.ns,"",it->cond_expr)
		<< std::endl;
#endif

      if(goal_map.find(property_id)!=goal_map.end())
      {
	//	it->cond_literal = !it->guard_literal;

        if(it->source.pc->location.get_comment()!=config->final)
        {
          goal_map[property_id].conjuncts.push_back(
		not_exprt(implies_exprt(conjunction(assumptions),
					it->cond_expr)));
        }
        else //final
        {
	  goal_map[property_id].conjuncts.clear(); //only keep last iteration
          goal_map[property_id].conjuncts.push_back(
		not_exprt(implies_exprt(conjunction(assumptions),
					it->cond_expr)));
        }
      }
    }
  }

  //check whether there are unreachable goals
  for(goal_mapt::const_iterator
      it=goal_map.begin();
      it!=goal_map.end();
      it++) 
  {
    if(it->second.conjuncts.size()==0) return false;
  }

  //set activation literal for incremental checking
  bmc.equation.new_activation_literal(bmc.symex.prop_conv);
  literalt activation_literal = bmc.equation.current_activation_literal();

  //add constraints /\_i \/_k -A_i^k for property_id i and iteration k  
  for(goal_mapt::const_iterator
      it=goal_map.begin();
      it!=goal_map.end();
      it++)
  {
    exprt::operandst conjuncts = it->second.conjuncts;
    conjuncts.push_back(literal_exprt(activation_literal));

#if 0
   std::cout << "goal " << it->first << ": "; 
    for(exprt::operandst::iterator it = conjuncts.begin();
        it!=conjuncts.end();it++) 
      std::cout << to_literal_expr(*it).get_literal() << " ";
    std::cout << std::endl;
#endif
 
    bmc.symex.prop_conv.set_to_true(disjunction(conjuncts));
  }

  return true;
}
